# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import copy
import numpy as np
import pandas as pd
import warnings
from datetime import datetime, timedelta

from grid2op.dtypes import dt_int, dt_float, dt_bool
from grid2op.Exceptions import (
    IncorrectNumberOfElements,
    ChronicsError,
    ChronicsNotFoundError,
)
from grid2op.Exceptions import (
    IncorrectNumberOfLoads,
    IncorrectNumberOfGenerators,
    IncorrectNumberOfLines,
)
from grid2op.Exceptions import EnvError, InsufficientData
from grid2op.Chronics.gridValue import GridValue


class GridStateFromFile(GridValue):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Do not attempt to create an object of this class. This is initialized by the environment
        at its creation.

    Read the injections values from a file stored on hard drive. More detailed about the files is provided in the
    :func:`GridStateFromFile.initialize` method.

    This class reads only files stored as csv. The header of the csv is mandatory and should represent the name of
    the objects. This names should either be matched to the name of the same object in the backend using the
    `names_chronics_to_backend` argument pass into the :func:`GridStateFromFile.initialize` (see
    :func:`GridValue.initialize` for more information) or match the names of the object in the backend.

    When the grid value is initialized, all present csv are read, sorted in order compatible with the backend and
    extracted as numpy array.

    For now, the current date and times are not read from file. It is mandatory that the chronics starts at 00:00 and
    its first time stamps is corresponds to January, 1st 2019.

    Chronics read from this files don't implement the "forecast" value.

    In this values, only 1 episode is stored. If the end of the episode is reached and another one should start, then
    it will loop from the beginning.

    It reads the following files from the "path" location specified:

    - "prod_p.csv": for each time steps, this file contains the value for the active production of
      each generators of the grid (it counts as many rows as the number of time steps - and its header)
      and as many columns as the number of generators on the grid. The header must contains the names of
      the generators used to map their value on the grid. Values must be convertible to floating point and the
      column separator of this file should be semi-colon `;` (unless you specify a "sep" when loading this class)
    - "prod_v.csv": same as "prod_p.csv" but for the production voltage setpoint.
    - "load_p.csv": same as "prod_p.csv" but for the load active value (number of columns = number of loads)
    - "load_q.csv": same as "prod_p.csv" but for the load reactive value (number of columns = number of loads)
    - "maintenance.csv": that contains whether or not there is a maintenance for a given powerline (column) at
      each time step (row).
    - "hazards.csv": that contains whether or not there is a hazard for a given powerline (column) at
      each time step (row).
    - "start_datetime.info": the time stamp (date and time) at which the chronic is starting.
    - "time_interval.info": the amount of time between two consecutive steps (*e.g.* 5 mins, or 1h)

    If a file is missing, it is understood as "this value will not be modified". For example, if the file
    "prod_v.csv" is not present, it will be equivalent as not modifying the production voltage setpoint, never.

    Except if the attribute :attr:`GridStateFromFile.sep` is modified, the above tables should be "semi colon" (;)
    separated.

    Attributes
    ----------
    path: ``str``
        The path of the folder where the data are stored. It is recommended to set absolute path, and not relative
        paths.

    load_p: ``numpy.ndarray``, dtype: ``float``
        All the values of the load active values

    load_q: ``numpy.ndarray``, dtype: ``float``
        All the values of the load reactive values

    prod_p: ``numpy.ndarray``, dtype: ``float``
        All the productions setpoint active values.

    prod_v: ``numpy.ndarray``, dtype: ``float``
        All the productions setpoint voltage magnitude values.

    hazards: ``numpy.ndarray``, dtype: ``bool``
        This vector represents the possible hazards. It is understood as: ``True`` there is a hazard
        for the given powerline, ``False`` there is not.

    maintenance: ``numpy.ndarray``, dtype: ``bool``
        This vector represents the possible maintenance. It is understood as: ``True`` there is a maintenance
        for the given powerline, ``False`` there is not.

    current_index: ``int``
        The index of the last observation sent to the :class:`grid2op.Environment`.

    sep: ``str``, optional
        The csv columns separator. By defaults it's ";"

    names_chronics_to_backend: ``dict``
        This directory matches the name of the objects (line extremity, generator or load) to the same object in the
        backed. See the help of :func:`GridValue.initialize` for more information).
    """
    MULTI_CHRONICS = False

    def __init__(
        self,
        path,
        sep=";",
        time_interval=timedelta(minutes=5),
        max_iter=-1,
        start_datetime=datetime(year=2019, month=1, day=1),
        chunk_size=None,
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Do not attempt to create an object of this class. This is initialized by the environment
        at its creation.


        Build an instance of GridStateFromFile. Such an instance should be built before an :class:`grid2op.Environment`
        is created.

        Parameters
        ----------
        path: ``str``
            Used to initialize :attr:`GridStateFromFile.path`

        sep: ``str``, optional
            Used to initialize :attr:`GridStateFromFile.sep`

        time_interval: ``datetime.timedelta``
            Used to initialize :attr:`GridValue.time_interval`

        max_iter: int, optional
            Used to initialize :attr:`GridValue.max_iter`

        """
        GridValue.__init__(
            self,
            time_interval=time_interval,
            max_iter=max_iter,
            start_datetime=start_datetime,
            chunk_size=chunk_size,
        )

        self.path = path
        self.n_ = None  # maximum number of rows of the array
        self.tmp_max_index = None  # size maximum of the current tables in memory
        self.load_p = None  # numpy array corresponding to the current active load values in the power _grid. It has the same size as the number of loads
        self.load_q = None  # numpy array corresponding to the current reactive load values in the power _grid. It has the same size as the number of loads
        self.prod_p = None  # numpy array corresponding to the current active production values in the power _grid. It has the same size as the number of generators
        self.prod_v = None  # numpy array corresponding to the current voltage production setpoint values in the power _grid. It has the same size as the number of generators

        # for the two following vector, the convention is the following: False(line is disconnected) / True(line is connected)
        self.hazards = None  # numpy array representing the outage (unplanned), same size as the number of powerlines on the _grid.
        self.maintenance = None  # numpy array representing the _maintenance (planned withdrawal of a powerline), same size as the number of powerlines on the _grid.
        self.maintenance_time = None
        self.maintenance_duration = None

        self.current_index = -1
        self.sep = sep

        self.names_chronics_to_backend = None

        # added to provide an easier access to read data in chunk
        self.chunk_size = chunk_size
        self._data_chunk = {}
        self._order_load_p = None
        self._order_load_q = None
        self._order_prod_p = None
        self._order_prod_v = None
        self._order_hazards = None
        self._order_maintenance = None

        # order of the names in the backend
        self._order_backend_loads = None
        self._order_backend_prods = None
        self._order_backend_lines = None

    def _clear(self):        
        self.n_ = None  # maximum number of rows of the array
        self.tmp_max_index = None  # size maximum of the current tables in memory
        self.load_p = None  # numpy array corresponding to the current active load values in the power _grid. It has the same size as the number of loads
        self.load_q = None  # numpy array corresponding to the current reactive load values in the power _grid. It has the same size as the number of loads
        self.prod_p = None  # numpy array corresponding to the current active production values in the power _grid. It has the same size as the number of generators
        self.prod_v = None  # numpy array corresponding to the current voltage production setpoint values in the power _grid. It has the same size as the number of generators

        # for the two following vector, the convention is the following: False(line is disconnected) / True(line is connected)
        self.hazards = None  # numpy array representing the outage (unplanned), same size as the number of powerlines on the _grid.
        self.maintenance = None  # numpy array representing the _maintenance (planned withdrawal of a powerline), same size as the number of powerlines on the _grid.
        self.maintenance_time = None
        self.maintenance_duration = None

        self.current_index = -1

        self.names_chronics_to_backend = None

        # added to provide an easier access to read data in chunk
        self._data_chunk = {}
        self._order_load_p = None
        self._order_load_q = None
        self._order_prod_p = None
        self._order_prod_v = None
        self._order_hazards = None
        self._order_maintenance = None

        # order of the names in the backend
        self._order_backend_loads = None
        self._order_backend_prods = None
        self._order_backend_lines = None
        
    def _assert_correct(self, dict_convert, order_backend):
        len_backend = len(order_backend)
        len_dict_keys = len(dict_convert)
        vals = set(dict_convert.values())
        lend_dict_values = len(vals)

        if len_dict_keys != len_backend:
            err_msg = "Conversion mismatch between backend data {} elements and converter data {} (keys)"
            raise IncorrectNumberOfElements(err_msg.format(len_backend, len_dict_keys))
        if lend_dict_values != len_backend:
            err_msg = "Conversion mismatch between backend data {} elements and converter data {} (values)"
            raise IncorrectNumberOfElements(
                err_msg.format(len_backend, lend_dict_values)
            )

        for el in order_backend:
            if not el in vals:
                raise ChronicsError(
                    'Impossible to find element "{}" in the original converter data'.format(
                        el
                    )
                )

    def _assert_correct_second_stage(self, pandas_name, dict_convert, key, extra=""):
        for i, el in enumerate(pandas_name):
            if not el in dict_convert[key]:
                raise ChronicsError(
                    "Element named {} is found in the data (column {}) but it is not found on the "
                    'powergrid for data of type "{}".\nData in files  are: {}\n'
                    "Converter data are: {}".format(
                        el,
                        i + 1,
                        key,
                        sorted(list(pandas_name)),
                        sorted(list(dict_convert[key].keys())),
                    )
                )

    def _init_date_time(self):
        if os.path.exists(os.path.join(self.path, "start_datetime.info")):
            with open(os.path.join(self.path, "start_datetime.info"), "r") as f:
                a = f.read().rstrip().lstrip()
            try:
                tmp = datetime.strptime(a, "%Y-%m-%d %H:%M")
            except ValueError:
                tmp = datetime.strptime(a, "%Y-%m-%d")
            except Exception:
                raise ChronicsNotFoundError(
                    'Impossible to understand the content of "start_datetime.info". Make sure '
                    'it\'s composed of only one line with a datetime in the "%Y-%m-%d %H:%M"'
                    "format."
                )
            self.start_datetime = tmp
            self.current_datetime = tmp

        if os.path.exists(os.path.join(self.path, "time_interval.info")):
            with open(os.path.join(self.path, "time_interval.info"), "r") as f:
                a = f.read().rstrip().lstrip()
            try:
                tmp = datetime.strptime(a, "%H:%M")
            except ValueError:
                tmp = datetime.strptime(a, "%M")
            except Exception:
                raise ChronicsNotFoundError(
                    'Impossible to understand the content of "time_interval.info". Make sure '
                    'it\'s composed of only one line with a datetime in the "%H:%M"'
                    "format."
                )
            self.time_interval = timedelta(hours=tmp.hour, minutes=tmp.minute)

    def _get_fileext(self, data_name):
        read_compressed = ".csv"
        if not os.path.exists(os.path.join(self.path, "{}.csv".format(data_name))):
            # try to read compressed data
            if os.path.exists(os.path.join(self.path, "{}.csv.bz2".format(data_name))):
                read_compressed = ".csv.bz2"
            elif os.path.exists(os.path.join(self.path, "{}.zip".format(data_name))):
                read_compressed = ".zip"
            elif os.path.exists(
                os.path.join(self.path, "{}.csv.gzip".format(data_name))
            ):
                read_compressed = ".csv.gzip"
            elif os.path.exists(os.path.join(self.path, "{}.csv.xz".format(data_name))):
                read_compressed = ".csv.xz"
            else:
                read_compressed = None
                # raise ChronicsNotFoundError(
                #     "GridStateFromFile: unable to locate the data files that should be at \"{}\"".format(self.path))
        return read_compressed

    def _get_data(self, data_name, chunksize=-1, nrows=None):
        file_ext = self._get_fileext(data_name)
        
        if nrows is None:
            if self._max_iter > 0:
                nrows = self._max_iter + 1
            
        if file_ext is not None:
            if chunksize == -1:
                chunksize = self.chunk_size
            res = pd.read_csv(
                os.path.join(self.path, "{}{}".format(data_name, file_ext)),
                sep=self.sep,
                chunksize=chunksize,
                nrows=nrows,
            )
        else:
            res = None
        return res

    def _get_orders(
        self,
        load_p,
        load_q,
        prod_p,
        prod_v,
        hazards,
        maintenance,
        order_backend_loads,
        order_backend_prods,
        order_backend_lines,
    ):

        order_chronics_load_p = None
        order_backend_load_q = None
        order_backend_prod_p = None
        order_backend_prod_v = None
        order_backend_hazards = None
        order_backend_maintenance = None

        if load_p is not None:
            self._assert_correct_second_stage(
                load_p.columns, self.names_chronics_to_backend, "loads", "active"
            )
            order_chronics_load_p = np.array(
                [
                    order_backend_loads[self.names_chronics_to_backend["loads"][el]]
                    for el in load_p.columns
                ]
            ).astype(dt_int)
        if load_q is not None:
            self._assert_correct_second_stage(
                load_q.columns, self.names_chronics_to_backend, "loads", "reactive"
            )
            order_backend_load_q = np.array(
                [
                    order_backend_loads[self.names_chronics_to_backend["loads"][el]]
                    for el in load_q.columns
                ]
            ).astype(dt_int)

        if prod_p is not None:
            self._assert_correct_second_stage(
                prod_p.columns, self.names_chronics_to_backend, "prods", "active"
            )
            order_backend_prod_p = np.array(
                [
                    order_backend_prods[self.names_chronics_to_backend["prods"][el]]
                    for el in prod_p.columns
                ]
            ).astype(dt_int)

        if prod_v is not None:
            self._assert_correct_second_stage(
                prod_v.columns,
                self.names_chronics_to_backend,
                "prods",
                "voltage magnitude",
            )
            order_backend_prod_v = np.array(
                [
                    order_backend_prods[self.names_chronics_to_backend["prods"][el]]
                    for el in prod_v.columns
                ]
            ).astype(dt_int)

        if hazards is not None:
            self._assert_correct_second_stage(
                hazards.columns, self.names_chronics_to_backend, "lines", "hazards"
            )
            order_backend_hazards = np.array(
                [
                    order_backend_lines[self.names_chronics_to_backend["lines"][el]]
                    for el in hazards.columns
                ]
            ).astype(dt_int)

        if maintenance is not None:
            self._assert_correct_second_stage(
                maintenance.columns,
                self.names_chronics_to_backend,
                "lines",
                "maintenance",
            )
            order_backend_maintenance = np.array(
                [
                    order_backend_lines[self.names_chronics_to_backend["lines"][el]]
                    for el in maintenance.columns
                ]
            ).astype(dt_int)

        return (
            order_chronics_load_p,
            order_backend_load_q,
            order_backend_prod_p,
            order_backend_prod_v,
            order_backend_hazards,
            order_backend_maintenance,
        )

    def _get_next_chunk(self):
        load_p = None
        load_q = None
        prod_p = None
        prod_v = None
        if self._data_chunk["load_p"] is not None:
            load_p = next(self._data_chunk["load_p"])
            self.tmp_max_index = load_p.shape[0]
        if self._data_chunk["load_q"] is not None:
            load_q = next(self._data_chunk["load_q"])
            self.tmp_max_index = load_q.shape[0]
        if self._data_chunk["prod_p"] is not None:
            prod_p = next(self._data_chunk["prod_p"])
            self.tmp_max_index = prod_p.shape[0]
        if self._data_chunk["prod_v"] is not None:
            prod_v = next(self._data_chunk["prod_v"])
            self.tmp_max_index = prod_v.shape[0]
        return load_p, load_q, prod_p, prod_v

    def initialize(
        self,
        order_backend_loads,
        order_backend_prods,
        order_backend_lines,
        order_backend_subs,
        names_chronics_to_backend=None,
    ):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Called at the creation of the environment.

        In this function, the numpy arrays are read from the csv using the panda.dataframe engine.

        In order to be valid, the folder located at :attr:`GridStateFromFile.path` can contain:

          - a file named "load_p.csv" used to initialize :attr:`GridStateFromFile.load_p`
          - a file named "load_q.csv" used to initialize :attr:`GridStateFromFile.load_q`
          - a file named "prod_p.csv" used to initialize :attr:`GridStateFromFile.prod_p`
          - a file named "prod_v.csv" used to initialize :attr:`GridStateFromFile.prod_v`
          - a file named "hazards.csv" used to initialize :attr:`GridStateFromFile.hazards`
          - a file named "maintenance.csv" used to initialize :attr:`GridStateFromFile.maintenance`

        All these csv must have the same separator specified by :attr:`GridStateFromFile.sep`.
        If one of these file is missing, it is equivalent to "change nothing" class.

        If a file named "start_datetime.info" is present, then it will be used to initialized
        :attr:`GridStateFromFile.start_datetime`. If this file exists, it should count only one row, with the
        initial datetime in the "%Y-%m-%d %H:%M" format.

        If a file named "time_interval.info" is present, then it will be used to initialized the
        :attr:`GridStateFromFile.time_interval` attribute.  If this file exists, it should count only one row, with the
        initial datetime in the "%H:%M" format. Only timedelta composed of hours and minutes are supported (time delta
        cannot go above 23 hours 55 minutes and cannot be smaller than 0 hour 1 minutes)

        The first row of these csv is understood as the name of the object concerned by the column. Either this name is
        present in the :class:`grid2op.Backend`, in this case no modification is performed, or in case the name
        is not found in the backend and in this case it must be specified in the "names_chronics_to_backend"
        parameters how to understand it. See the help of :func:`GridValue.initialize` for more information
        about this dictionnary.

        All files should have the same number of rows.

        Parameters
        ----------
        See help of :func:`GridValue.initialize` for a detailed help about the parameters.

        """
        self.n_gen = len(order_backend_prods)
        self.n_load = len(order_backend_loads)
        self.n_line = len(order_backend_lines)

        self._order_backend_loads = order_backend_loads
        self._order_backend_prods = order_backend_prods
        self._order_backend_lines = order_backend_lines

        self.names_chronics_to_backend = copy.deepcopy(names_chronics_to_backend)
        if self.names_chronics_to_backend is None:
            self.names_chronics_to_backend = {}
        if not "loads" in self.names_chronics_to_backend:
            self.names_chronics_to_backend["loads"] = {
                k: k for k in order_backend_loads
            }
        else:
            self._assert_correct(
                self.names_chronics_to_backend["loads"], order_backend_loads
            )
        if not "prods" in self.names_chronics_to_backend:
            self.names_chronics_to_backend["prods"] = {
                k: k for k in order_backend_prods
            }
        else:
            self._assert_correct(
                self.names_chronics_to_backend["prods"], order_backend_prods
            )
        if not "lines" in self.names_chronics_to_backend:
            self.names_chronics_to_backend["lines"] = {
                k: k for k in order_backend_lines
            }
        else:
            self._assert_correct(
                self.names_chronics_to_backend["lines"], order_backend_lines
            )
        if not "subs" in self.names_chronics_to_backend:
            self.names_chronics_to_backend["subs"] = {k: k for k in order_backend_subs}
        else:
            self._assert_correct(
                self.names_chronics_to_backend["subs"], order_backend_subs
            )

        self._init_date_time()

        # read the data
        load_p_iter = self._get_data("load_p")
        load_q_iter = self._get_data("load_q")
        prod_p_iter = self._get_data("prod_p")
        prod_v_iter = self._get_data("prod_v")
        read_compressed = self._get_fileext("hazards")
        nrows = None
        if self._max_iter > 0:
            nrows = self._max_iter + 1

        if read_compressed is not None:
            hazards = pd.read_csv(
                os.path.join(self.path, "hazards{}".format(read_compressed)),
                sep=self.sep,
                nrows=nrows,
            )
        else:
            hazards = None

        read_compressed = self._get_fileext("maintenance")
        if read_compressed is not None:
            maintenance = pd.read_csv(
                os.path.join(self.path, "maintenance{}".format(read_compressed)),
                sep=self.sep,
                nrows=nrows,
            )
        else:
            maintenance = None

        # put the proper name in order
        order_backend_loads = {el: i for i, el in enumerate(order_backend_loads)}
        order_backend_prods = {el: i for i, el in enumerate(order_backend_prods)}
        order_backend_lines = {el: i for i, el in enumerate(order_backend_lines)}

        if self.chunk_size is None:
            load_p = load_p_iter
            load_q = load_q_iter
            prod_p = prod_p_iter
            prod_v = prod_v_iter
            if load_p is not None:
                self.tmp_max_index = load_p.shape[0]
            elif load_q is not None:
                self.tmp_max_index = load_q.shape[0]
            elif prod_p is not None:
                self.tmp_max_index = prod_p.shape[0]
            elif prod_v is not None:
                self.tmp_max_index = prod_v.shape[0]
            else:
                raise ChronicsError(
                    'No files are found in directory "{}". If you don\'t want to load any chronics,'
                    ' use  "ChangeNothing" and not "{}" to load chronics.'
                    "".format(self.path, type(self))
                )

        else:
            self._data_chunk = {
                "load_p": load_p_iter,
                "load_q": load_q_iter,
                "prod_p": prod_p_iter,
                "prod_v": prod_v_iter,
            }
            load_p, load_q, prod_p, prod_v = self._get_next_chunk()

        # get the chronics in order
        (
            order_chronics_load_p,
            order_backend_load_q,
            order_backend_prod_p,
            order_backend_prod_v,
            order_backend_hazards,
            order_backend_maintenance,
        ) = self._get_orders(
            load_p,
            load_q,
            prod_p,
            prod_v,
            hazards,
            maintenance,
            order_backend_loads,
            order_backend_prods,
            order_backend_lines,
        )

        # now "sort" the columns of each chunk of data
        self._order_load_p = np.argsort(order_chronics_load_p)
        self._order_load_q = np.argsort(order_backend_load_q)
        self._order_prod_p = np.argsort(order_backend_prod_p)
        self._order_prod_v = np.argsort(order_backend_prod_v)
        self._order_hazards = np.argsort(order_backend_hazards)
        self._order_maintenance = np.argsort(order_backend_maintenance)

        # retrieve total number of rows
        if maintenance is not None:
            n_ = maintenance.shape[0]
        elif hazards is not None:
            n_ = hazards.shape[0]
        else:
            n_ = None
            for fn in ["prod_p", "load_p", "prod_v", "load_q"]:
                ext_ = self._get_fileext(fn)
                if ext_ is not None:
                    n_ = self._file_len(
                        os.path.join(self.path, "{}{}".format(fn, ext_)), ext_
                    )
                    break
            if n_ is None:
                raise ChronicsError(
                    'No files are found in directory "{}". If you don\'t want to load any chronics,'
                    ' use  "ChangeNothing" and not "{}" to load chronics.'
                    "".format(self.path, type(self))
                )
        self.n_ = n_  # the -1 is present because the initial grid state doesn't count as a "time step"

        if self._max_iter > 0:
            if self.n_ is not None:
                if self._max_iter >= self.n_:
                    self._max_iter = self.n_ - 1
                    # TODO: issue warning in this case
            self.n_ = self._max_iter + 1
        else:
            # if the number of maximum time step is not set yet, we set it to be the number of
            # data in the chronics (number of rows of the files) -1.
            # the -1 is present because the initial grid state doesn't count as a "time step" but is read
            # from these data.
            self._max_iter = self.n_ - 1

        self._init_attrs(
            load_p, load_q, prod_p, prod_v, hazards=hazards, maintenance=maintenance,
            is_init=True
        )

        self.curr_iter = 0

    @staticmethod
    def _file_len(fname, ext_):
        res = pd.read_csv(fname, sep="@", dtype=str).shape[0]
        return res

    def _init_attrs(
        self, load_p, load_q, prod_p, prod_v, hazards=None, maintenance=None,
        is_init=False
    ):
        # this called at the initialization but also each time more data should
        # be read from the disk (at the end of each chunk for example)
        self.load_p = None
        self.load_q = None
        self.prod_p = None
        self.prod_v = None
        if is_init:
            self.hazards = None
            self.hazard_duration = None
            self.maintenance = None
            self.maintenance_time = None
            self.maintenance_duration = None

        if load_p is not None:
            self.load_p = copy.deepcopy(
                load_p.values[:, self._order_load_p].astype(dt_float)
            )
        if load_q is not None:
            self.load_q = copy.deepcopy(
                load_q.values[:, self._order_load_q].astype(dt_float)
            )
        if prod_p is not None:
            self.prod_p = copy.deepcopy(
                prod_p.values[:, self._order_prod_p].astype(dt_float)
            )
        if prod_v is not None:
            self.prod_v = copy.deepcopy(
                prod_v.values[:, self._order_prod_v].astype(dt_float)
            )

        # TODO optimize this piece of code, and the whole laoding process if hazards.csv and maintenance.csv are
        # provided in the proper format.
        if hazards is not None:
            # hazards and maintenance cannot be computed by chunk. So we need to differenciate their behaviour
            self.hazards = copy.deepcopy(hazards.values[:, self._order_hazards])
            self.hazard_duration = np.zeros(
                shape=(self.hazards.shape[0], self.n_line), dtype=dt_int
            )
            for line_id in range(self.n_line):
                self.hazard_duration[:, line_id] = self.get_hazard_duration_1d(
                    self.hazards[:, line_id]
                )

            self.hazards = self.hazards != 0.0
        if maintenance is not None:
            self.maintenance = copy.deepcopy(
                maintenance.values[:, self._order_maintenance]
            )
            self.maintenance_time = (
                np.zeros(shape=(self.maintenance.shape[0], self.n_line), dtype=dt_int)
                - 1
            )
            self.maintenance_duration = np.zeros(
                shape=(self.maintenance.shape[0], self.n_line), dtype=dt_int
            )

            # test that with chunk size
            for line_id in range(self.n_line):
                self.maintenance_time[:, line_id] = self.get_maintenance_time_1d(
                    self.maintenance[:, line_id]
                )
                self.maintenance_duration[
                    :, line_id
                ] = self.get_maintenance_duration_1d(self.maintenance[:, line_id])

            # there are _maintenance and hazards only if the value in the file is not 0.
            self.maintenance = self.maintenance != 0.0
            self.maintenance = self.maintenance.astype(dt_bool)

    def done(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

        Compare to :func:`GridValue.done` an episode can be over for 2 main reasons:

          - :attr:`GridValue.max_iter` has been reached
          - There are no data in the csv.

        The episode is done if one of the above condition is met.

        Returns
        -------
        res: ``bool``
            Whether the episode has reached its end or not.

        """
        res = False
        # if self.current_index+1 >= self.tmp_max_index:
        if self.current_index > self.n_:
            res = True
        elif self._max_iter > 0:
            if self.curr_iter > self._max_iter:
                res = True
        return res

    @property
    def max_iter(self):
        return self._max_iter
    
    @max_iter.setter
    def max_iter(self, value : int):
        if value == -1:
            self._max_iter = self.n_ - 1
        else:
            self._max_iter = int(value)
        
    def max_timestep(self):
        if self._max_iter == -1:
            return self.n_ - 1
        return self._max_iter 
    
    def _data_in_memory(self):
        if self.chunk_size is None:
            # if i don't use chunk, all the data are in memory alreay
            return True
        if self.current_index == 0:
            # data are loaded the first iteration
            return True
        if self.current_index % self.chunk_size != 0:
            # data are already in ram
            return True
        return False

    def _load_next_chunk_in_memory(self):
        # print("I loaded another chunk")
        # i load the next chunk as dataframes
        load_p, load_q, prod_p, prod_v = self._get_next_chunk()
        # i put these dataframes in the right order (columns)
        self._init_attrs(load_p, load_q, prod_p, prod_v)
        # i don't forget to reset the reading index to 0
        self.current_index = 0

    def load_next(self):
        self.current_index += 1  # index in the chunk
        # for the "global" index use self.curr_iter

        if not self._data_in_memory():
            try:
                self._load_next_chunk_in_memory()
            except StopIteration as exc_:
                raise StopIteration from exc_

        if self.current_index >= self.tmp_max_index:
            raise StopIteration

        if self._max_iter > 0:
            if self.curr_iter > self._max_iter:
                raise StopIteration

        res = {}
        dict_ = {}
        prod_v = None
        if self.load_p is not None:
            dict_["load_p"] = 1.0 * self.load_p[self.current_index, :]
        if self.load_q is not None:
            dict_["load_q"] = 1.0 * self.load_q[self.current_index, :]
        if self.prod_p is not None:
            dict_["prod_p"] = 1.0 * self.prod_p[self.current_index, :]
        if self.prod_v is not None:
            prod_v = 1.0 * self.prod_v[self.current_index, :]
        if dict_:
            res["injection"] = dict_

        if self.maintenance is not None:
            res["maintenance"] = self.maintenance[self.curr_iter, :]
        if self.hazards is not None:
            res["hazards"] = self.hazards[self.curr_iter, :]

        if self.maintenance_time is not None:
            maintenance_time = dt_int(1 * self.maintenance_time[self.curr_iter, :])
            maintenance_duration = dt_int(
                1 * self.maintenance_duration[self.curr_iter, :]
            )
        else:
            maintenance_time = np.full(self.n_line, fill_value=-1, dtype=dt_int)
            maintenance_duration = np.full(self.n_line, fill_value=0, dtype=dt_int)

        if self.hazard_duration is not None:
            hazard_duration = 1 * self.hazard_duration[self.current_index, :]
        else:
            hazard_duration = np.full(self.n_line, fill_value=-1, dtype=dt_int)

        self.current_datetime += self.time_interval
        self.curr_iter += 1
        return (
            self.current_datetime,
            res,
            maintenance_time,
            maintenance_duration,
            hazard_duration,
            prod_v,
        )

    def check_validity(self, backend):
        at_least_one = False
        if self.load_p is not None:
            if self.load_p.shape[1] != backend.n_load:
                msg_err = "for the active part. It should be {} but is in fact {}"
                raise IncorrectNumberOfLoads(
                    msg_err.format(backend.n_load, self.load_p.shape[1])
                )
            at_least_one = True

        if self.load_q is not None:
            if self.load_q.shape[1] != backend.n_load:
                msg_err = "for the reactive part. It should be {} but is in fact {}"
                raise IncorrectNumberOfLoads(
                    msg_err.format(backend.n_load, self.load_q.shape[1])
                )
            at_least_one = True
        if self.prod_p is not None:
            if self.prod_p.shape[1] != backend.n_gen:
                msg_err = "for the active part. It should be {} but is in fact {}"
                raise IncorrectNumberOfGenerators(
                    msg_err.format(backend.n_gen, self.prod_p.shape[1])
                )
            at_least_one = True

        if self.prod_v is not None:
            if self.prod_v.shape[1] != backend.n_gen:
                msg_err = "for the voltage part. It should be {} but is in fact {}"
                raise IncorrectNumberOfGenerators(
                    msg_err.format(backend.n_gen, self.prod_v.shape[1])
                )
            at_least_one = True

        if self.hazards is not None:
            if self.hazards.shape[1] != backend.n_line:
                msg_err = "for the outage. It should be {} but is in fact {}"
                raise IncorrectNumberOfLines(
                    msg_err.format(backend.n_line, self.hazards.shape[1])
                )
            at_least_one = True

        if self.maintenance is not None:
            if self.maintenance.shape[1] != backend.n_line:
                msg_err = "for the maintenance. It should be {} but is in fact {}"
                raise IncorrectNumberOfLines(
                    msg_err.format(backend.n_line, self.maintenance.shape[1])
                )
            at_least_one = True

        if self.maintenance_time is not None:
            if self.maintenance_time.shape[1] != backend.n_line:
                msg_err = "for the maintenance times. It should be {} but is in fact {}"
                raise IncorrectNumberOfLines(
                    msg_err.format(backend.n_line, self.maintenance_time.shape[1])
                )
            at_least_one = True

        if self.maintenance_duration is not None:
            if self.maintenance_duration.shape[1] != backend.n_line:
                msg_err = (
                    "for the maintenance durations. It should be {} but is in fact {}"
                )
                raise IncorrectNumberOfLines(
                    msg_err.format(backend.n_line, self.maintenance_duration.shape[1])
                )
            at_least_one = True

        if self.hazard_duration is not None:
            if self.hazard_duration.shape[1] != backend.n_line:
                msg_err = "for the hazard durations. It should be {} but is in fact {}"
                raise IncorrectNumberOfLines(
                    msg_err.format(backend.n_line, self.hazard_duration.shape[1])
                )
            at_least_one = True

        if not at_least_one:
            raise ChronicsError(
                'No files are found in directory "{}". If you don\'t want to load any chronics, use '
                '"ChangeNothing" and not "{}" to load chronics.'
                "".format(self.path, type(self))
            )

        for name_arr, arr in zip(
            [
                "load_q",
                "load_p",
                "prod_v",
                "prod_p",
                "maintenance",
                "hazards",
                "maintenance time",
                "maintenance duration",
                "hazard duration",
            ],
            [
                self.load_q,
                self.load_p,
                self.prod_v,
                self.prod_p,
                self.maintenance,
                self.hazards,
                self.maintenance_time,
                self.maintenance_duration,
                self.hazard_duration,
            ],
        ):
            if arr is not None:
                if self.chunk_size is None:
                    if arr.shape[0] != self.n_:
                        msg_err = (
                            "Array {} has not the same number of rows ({}) than the maintenance ({}). "
                            "The chronics cannot be loaded properly."
                        )
                        raise EnvError(msg_err.format(name_arr, arr.shape[0], self.n_))

        if self._max_iter > 0:
            if self._max_iter > self.n_:
                msg_err = "Files count {} rows and you ask this episode to last at {} timestep."
                raise InsufficientData(msg_err.format(self.n_, self._max_iter))

    def next_chronics(self):
        self.current_datetime = self.start_datetime
        self.current_index = -1
        self.curr_iter = 0
        if self.chunk_size is not None:
            self._clear()  # remove previously loaded data [only needed if chunk size is set, I assume]

    def get_id(self) -> str:
        return self.path

    def set_chunk_size(self, new_chunk_size):
        self.chunk_size = new_chunk_size

    def _convert_datetime(self, datetime_beg):
        res = datetime_beg
        if not isinstance(datetime_beg, datetime):
            try:
                res = datetime.strptime(datetime_beg, "%Y-%m-%d %H:%M")
            except:
                try:
                    res = datetime.strptime(datetime_beg, "%Y-%m-%d")
                except:
                    raise ChronicsError(
                        'Impossible to convert "{}" to a valid datetime. Accepted format is '
                        '"%Y-%m-%d %H:%M"'.format(datetime_beg)
                    )
        return res

    def _extract_array(self, nm):
        var = self.__dict__[nm]
        if var is None:
            return None
        else:
            return var[self.current_index, :]

    def _save_array(self, array_, path_out, name, colnames):
        if array_ is None:
            return
        tmp = pd.DataFrame(array_)
        tmp.columns = colnames
        tmp.to_csv(os.path.join(path_out, name), index=False, sep=self.sep)

    def _init_res_split(self, nb_rows):
        res_prod_p = None
        res_prod_v = None
        res_load_p = None
        res_load_q = None
        res_maintenance = None
        res_hazards = None
        if self.prod_p is not None:
            res_prod_p = np.zeros((nb_rows, self.n_gen), dtype=dt_float)
        if self.prod_v is not None:
            res_prod_v = np.zeros((nb_rows, self.n_gen), dtype=dt_float)
        if self.load_p is not None:
            res_load_p = np.zeros((nb_rows, self.n_load), dtype=dt_float)
        if self.load_q is not None:
            res_load_q = np.zeros((nb_rows, self.n_load), dtype=dt_float)
        if self.maintenance is not None:
            res_maintenance = np.zeros((nb_rows, self.n_line), dtype=dt_float)
        if self.hazards is not None:
            res_hazards = np.zeros((nb_rows, self.n_line), dtype=dt_float)
        return (
            res_prod_p,
            res_prod_v,
            res_load_p,
            res_load_q,
            res_maintenance,
            res_hazards,
        )

    def _update_res_split(self, i, tmp, *arrays):
        (
            res_prod_p,
            res_prod_v,
            res_load_p,
            res_load_q,
            res_maintenance,
            res_hazards,
        ) = arrays
        if res_prod_p is not None:
            res_prod_p[i, :] = tmp._extract_array("prod_p")
        if res_prod_v is not None:
            res_prod_v[i, :] = tmp._extract_array("prod_v")
        if res_load_p is not None:
            res_load_p[i, :] = tmp._extract_array("load_p")
        if res_load_q is not None:
            res_load_q[i, :] = tmp._extract_array("load_q")
        if res_maintenance is not None:
            res_maintenance[i, :] = tmp._extract_array("maintenance")
        if res_hazards is not None:
            res_hazards[i, :] = tmp._extract_array("hazards")

    def _clean_arrays(self, i, *arrays):
        (
            res_prod_p,
            res_prod_v,
            res_load_p,
            res_load_q,
            res_maintenance,
            res_hazards,
        ) = arrays
        if res_prod_p is not None:
            res_prod_p = res_prod_p[:i, :]
        if res_prod_v is not None:
            res_prod_v = res_prod_v[:i, :]
        if res_load_p is not None:
            res_load_p = res_load_p[:i, :]
        if res_load_q is not None:
            res_load_q = res_load_q[:i, :]
        if res_maintenance is not None:
            res_maintenance = res_maintenance[:i, :]
        if res_hazards is not None:
            res_hazards = res_hazards[:i, :]
        return (
            res_prod_p,
            res_prod_v,
            res_load_p,
            res_load_q,
            res_maintenance,
            res_hazards,
        )

    def _get_name_arrays_for_saving(self):
        return ["prod_p", "prod_v", "load_p", "load_q", "maintenance", "hazards"]

    def _get_colorder_arrays_for_saving(self):
        return [
            self._order_backend_prods,
            self._order_backend_prods,
            self._order_backend_loads,
            self._order_backend_loads,
            self._order_backend_lines,
            self._order_backend_lines,
        ]

    def split_and_save(self, datetime_beg, datetime_end, path_out):
        """
        You can use this function to save the values of the chronics in a format that will be loadable
        by :class:`GridStateFromFile`

        Notes
        -----
        Prefer using the :func:`Multifolder.split_and_save` that handles different chronics

        Parameters
        ----------
        datetime_beg: ``str``
            Time stamp of the beginning of the data you want to save (time stamp in "%Y-%m-%d %H:%M"
            format)

        datetime_end: ``str``
            Time stamp of the end of the data you want to save (time stamp in "%Y-%m-%d %H:%M"
            format)

        path_out: ``str``
            Location where to save the data

        """
        # work on a copy of myself
        tmp = copy.deepcopy(self)
        datetime_beg = self._convert_datetime(datetime_beg)
        datetime_end = self._convert_datetime(datetime_end)

        nb_rows = datetime_end - datetime_beg
        nb_rows = nb_rows.total_seconds()
        nb_rows = int(nb_rows / self.time_interval.total_seconds()) + 1
        if nb_rows <= 0:
            raise ChronicsError(
                'Invalid time step to be extracted. Make sure "datetime_beg" is lower than '
                '"datetime_end" {} - {}'.format(datetime_beg, datetime_end)
            )

        # prepare folder
        if not os.path.exists(path_out):
            os.mkdir(path_out)

        # skip until datetime_beg starts
        curr_dt = tmp.current_datetime
        if curr_dt > datetime_beg:
            warnings.warn(
                "split_and_save: you ask for a beginning of the extraction of the chronics after the "
                "current datetime of it. If they ever existed, the data in the chronics prior to {}"
                "will be ignored".format(curr_dt)
            )
        # in the chronics we load the first row to initialize the data, so here we stop just a bit before that
        datetime_start = datetime_beg - self.time_interval
        while curr_dt < datetime_start:
            curr_dt, *_ = tmp.load_next()
        real_init_dt = curr_dt
        arrays = self._init_res_split(nb_rows)
        i = 0
        while curr_dt < datetime_end:
            self._update_res_split(i, tmp, *arrays)
            curr_dt, *_ = tmp.load_next()
            i += 1
        if i < nb_rows:
            warnings.warn(
                "split_and_save: chronics goes up to {} but you want to split it up to {}. Results "
                "has been troncated".format(curr_dt, datetime_end)
            )

        arrays = self._clean_arrays(i, *arrays)
        nms = self._get_name_arrays_for_saving()
        orders_columns = self._get_colorder_arrays_for_saving()
        for el, nm, colnames in zip(arrays, nms, orders_columns):
            nm = "{}{}".format(nm, ".csv.bz2")
            self._save_array(el, path_out, nm, colnames)

        with open(os.path.join(path_out, "start_datetime.info"), "w") as f:
            f.write("{:%Y-%m-%d %H:%M}\n".format(real_init_dt))

        tmp_for_time_delta = (
            datetime(year=2018, month=1, day=1, hour=0, minute=0, second=0)
            + self.time_interval
        )
        with open(os.path.join(path_out, "time_interval.info"), "w") as f:
            f.write("{:%H:%M}\n".format(tmp_for_time_delta))



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
from grid2op.Exceptions import Grid2OpException

# i already issued the warning for the "some substations have no controllable elements"
_WARNING_ISSUED_FOR_SUB_NO_ELEM = False
# this global variable is not const ! It is modified in GridObjects.py


def extract_from_dict(dict_, key, converter):
    if key not in dict_:
        raise Grid2OpException(
            'Impossible to find key "{}" while loading the dictionary.'.format(key)
        )
    try:
        res = converter(dict_[key])
    except Exception as exc_:
        raise Grid2OpException(
            'Impossible to convert "{}" into class {} with exception '
            '\n"{}"'.format(key, converter, exc_)
        )
    return res


def save_to_dict(res_dict, me, key, converter, copy_=True):
    """

    Parameters
    ----------
    res_dict:
        output dictionary
    me:
        the object to serialize in a dict
    key:
        the attribute of the object we want to save
    converter:
        if the attribute need to be converted (for example if you later want to serialize the dictionary as json)
    copy_:
        whether you copy the attribute or not (only applies if converter is None)

    Returns
    -------

    """
    if not hasattr(me, key):
        raise Grid2OpException(
            'Impossible to find key "{}" while loading the dictionary.'.format(key)
        )
    try:
        if converter is not None:
            res = converter(getattr(me, key))
        else:
            if copy_:
                res = copy.deepcopy(getattr(me, key))
            else:
                res = getattr(me, key)
    except Exception as exc_:
        raise Grid2OpException(
            'Impossible to convert "{}" into class {} with exception '
            '\n"{}"'.format(key, converter, exc_)
        )

    if key in res_dict:
        msg_err_ = (
            'Key "{}" is already present in the result dictionary. This would override it'
            " and is not supported."
        )
        raise Grid2OpException(msg_err_.format(key))
    res_dict[key] = res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import os
import json
import warnings
import numpy as np
from datetime import timedelta, datetime

from grid2op.dtypes import dt_int, dt_float
from grid2op.Exceptions import ChronicsNotFoundError, ChronicsError
from grid2op.Chronics.gridValue import GridValue
from grid2op.Chronics.gridStateFromFile import GridStateFromFile


class Multifolder(GridValue):
    """
    The classes :class:`GridStateFromFile` and :class:`GridStateFromFileWithForecasts` implemented the reading of a
    single folder representing a single episode.

    This class is here to "loop" between different episode. Each one being stored in a folder readable by
    :class:`GridStateFromFile` or one of its derivate (eg. :class:`GridStateFromFileWithForecasts`).

    Chronics are always read in the alpha-numeric order for this class. This means that if the folder is not modified,
    the data are always loaded in the same order, regardless of the :class:`grid2op.Backend`, :class:`grid2op.BaseAgent` or
    :class:`grid2op.Environment`.

    .. note::
        Most grid2op environments, by default, use this type of "chronix", read from the hard drive.
        
    Attributes
    -----------
    gridvalueClass: ``type``, optional
        Type of class used to read the data from the disk. It defaults to :class:`GridStateFromFile`.

    data: :class:`GridStateFromFile`
        Data that will be loaded and used to produced grid state and forecasted values.


    path: ``str``
        Path where the folders of the episodes are stored.

    sep: ``str``
        Columns separtor, forwarded to :attr:`Multifolder.data` when it's built at the beginning of each episode.

    subpaths: ``list``
        List of all the episode that can be "played". It's a sorted list of all the directory in
        :attr:`Multifolder.path`. Each one should contain data in a format that is readable by
        :attr:`MultiFolder.gridvalueClass`.

    """
    MULTI_CHRONICS = True

    def __init__(
        self,
        path,
        time_interval=timedelta(minutes=5),
        start_datetime=datetime(year=2019, month=1, day=1),
        gridvalueClass=GridStateFromFile,
        sep=";",
        max_iter=-1,
        chunk_size=None,
        filter_func=None,
        **kwargs
    ):
        self._kwargs = kwargs
        GridValue.__init__(
            self,
            time_interval=time_interval,
            max_iter=max_iter,
            chunk_size=chunk_size,
            start_datetime=start_datetime,
        )
        self.gridvalueClass = gridvalueClass
        self.data = None
        self.path = os.path.abspath(path)
        self.sep = sep
        self.init_subpath()

        if len(self.subpaths) == 0:
            raise ChronicsNotFoundError(
                'Not chronics are found in "{}". Make sure there are at least '
                "1 chronics folder there.".format(self.path)
            )
        # TODO clarify that
        # np.random.shuffle(self.subpaths)
        self.chunk_size = chunk_size

        # for saving
        self._order_backend_loads = None
        self._order_backend_prods = None
        self._order_backend_lines = None
        self._order_backend_subs = None
        self._names_chronics_to_backend = None

        # improving looping strategy
        if filter_func is None:
            self._filter = self._default_filter
        else:
            if not callable(filter_func):
                raise ChronicsError(
                    "The filtering function you provided ("
                    "kwargs: filter_func) is not callable."
                )
            self._filter = filter_func
        self._prev_cache_id = 0
        self._order = None

    def init_subpath(self):
        """
        Read the content of the main directory and initialize the `subpaths` 
        where the data could be located.
        
        This is usefull, for example, if you generated data and want to be able to use them.
        
        **NB** this has no effect until :attr:`Multifolder.reset` is called.
        
        .. warning::
            By default, it will only consider data that are present at creation time. If you add data after, you need
            to call this function (and do a reset)
            
        Examples
        ---------
        
        A "typical" usage of this function can be the following workflow.
        
        Start a script to train an agent (say "train_agent.py"):
        
        .. code-block:: python
        
            import os
            import grid2op
            from lightsim2grid import LightSimBackend  # highly recommended for speed !
            
            env_name = "l2rpn_wcci_2022"  # only compatible with what comes next (at time of writing)
            env = grid2op.make(env_name, backend=LightSimBackend())
            
            # now train an agent
            # see l2rpn_baselines package for more information, for example
            # l2rpn-baselines.readthedocs.io/
            from l2rpn_baselines.PPO_SB3 import train
            nb_iter = 10000  # train for that many iterations
            agent_name = "WhaetverIWant"  # or any other name
            agent_path = os.path.expand("~")  # or anywhere else on your computer
            trained_agent = train(env,
                                  iterations=nb_iter,
                                  name=agent_name,
                                  save_path=agent_path)
            
        On another script (say "generate_data.py"), you can generate more data:
                
        .. code-block:: python
        
            import grid2op            
            env_name = "l2rpn_wcci_2022"  # only compatible with what comes next (at time of writing)
            env = grid2op.make(env_name)
            env.generate_data(nb_year=50)  # generates 50 years of data 
            # (takes roughly 50s per week, around 45mins per year, in this case 50 * 45 mins = lots of minutes)
            
        Let the script to generate the data run normally (don't interupt it).
        And from time to time, in the script "train_agent.py" you can do:
        
        .. code-block:: python
            
            # reload the generated data
            env.chronics_handler.init_subpath()
            env.chronics_handler.reset()
            
            # retrain the agent taking into account new data
            trained_agent = train(env,
                                  iterations=nb_iter,
                                  name=agent_name,
                                  save_path=agent_path,
                                  load_path=agent_path
                                  )
            
            # the script to generate data is still running, you can reload some data again
            env.chronics_handler.init_subpath()
            env.chronics_handler.reset()
            
            # retrain the agent
            trained_agent = train(env,
                                  iterations=nb_iter,
                                  name=agent_name,
                                  save_path=agent_path,
                                  load_path=agent_path
                                  )
                                  
            # etc.
        
        Both scripts you run "at the same time" for it to work efficiently.
        
        To recap:
        - script "generate_data.py" will... generate data
        - these data will be reloaded from time to time by the script "train_agent.py"
        
        .. warning:: 
            Do not delete data between calls to `env.chronics_handler.init_subpath()` and `env.chronics_handler.reset()`,
            and even less so during training !
            
            If you want to delete data (for example not to overload your hard drive) you should remove them 
            right before calling `env.chronics_handler.init_subpath()`.
            
        """
        try:
            self.subpaths = [
                os.path.join(self.path, el)
                for el in os.listdir(self.path)
                if os.path.isdir(os.path.join(self.path, el)) and (el != "__pycache__") and (not el.startswith("."))
            ]
            self.subpaths.sort()
            self.subpaths = np.array(self.subpaths)
        except FileNotFoundError as exc_:
            raise ChronicsError(
                'Path "{}" doesn\'t exists.'.format(self.path)
            ) from exc_
        self._order = None  # to trigger a "reset" when chronix will next be loaded
        
    def get_kwargs(self, dict_):
        if self._filter != self._default_filter:
            dict_["filter_func"] = self._filter

    def available_chronics(self):
        """return the list of available chronics.

        Examples
        --------

        # TODO
        """
        return self.subpaths[self._order]

    def _default_filter(self, x):
        """
        default filter used at the initialization. It keeps only the first data encountered.
        """
        return True

    def set_filter(self, filter_fun):
        """
        Assign a filtering function to remove some chronics from the next time a call to "reset_cache" is called.

        **NB** filter_fun is applied to all element of :attr:`Multifolder.subpaths`. If ``True`` then it will
        be put in cache, if ``False`` this data will NOT be put in the cache.

        **NB** this has no effect until :attr:`Multifolder.reset` is called.

        Examples
        --------
        Let's assume in your chronics, the folder names are "Scenario_august_dummy", and
        "Scenario_february_dummy". For the sake of the example, we want the environment to loop
        only through the month of february, because why not. Then we can do the following:

        .. code-block:: python

            import re
            import grid2op
            env = grid2op.make("l2rpn_neurips_2020_track1", test=True)  # don't add "test=True" if
            # you don't want to perform a test.

            # check at which month will belong each observation
            for i in range(10):
                obs = env.reset()
                print(obs.month)
                # it always alternatively prints "8" (if chronics if from august) or
                # "2" if chronics is from february)

            # to see where the chronics are located
            print(env.chronics_handler.subpaths)

            # keep only the month of february
            env.chronics_handler.set_filter(lambda path: re.match(".*february.*", path) is not None)
            env.chronics_handler.reset()  # if you don't do that it will not have any effect

            for i in range(10):
                obs = env.reset()
                print(obs.month)
                # it always prints "2" (representing february)

        """
        self._filter = filter_fun

    def next_chronics(self):
        self._prev_cache_id += 1
        # TODO implement the shuffling indeed.
        # if self._prev_cache_id >= len(self._order):
        #     self.space_prng.shuffle(self._order)
        self._prev_cache_id %= len(self._order)

    def sample_next_chronics(self, probabilities=None):
        """
        This function should be called before "next_chronics".
        It can be used to sample non uniformly for the next next chronics.

        Parameters
        -----------
        probabilities: ``np.ndarray``
            Array of integer with the same size as the number of chronics in the cache.
            If it does not sum to one, it is rescaled such that it sums to one.

        Returns
        -------
        selected: ``int``
            The integer that was selected.

        Examples
        --------

        Let's assume in your chronics, the folder names are "Scenario_august_dummy", and
        "Scenario_february_dummy". For the sake of the example, we want the environment to loop
        75% of the time to the month of february and 25% of the time to the month of august.

        .. code-block:: python

            import grid2op
            env = grid2op.make("l2rpn_neurips_2020_track1", test=True)  # don't add "test=True" if
            # you don't want to perform a test.

            # check at which month will belong each observation
            for i in range(10):
                obs = env.reset()
                print(obs.month)
                # it always alternatively prints "8" (if chronics if from august) or
                # "2" if chronics is from february) with a probability of 50% / 50%

            env.seed(0)  # for reproducible experiment
            for i in range(10):
                _ = env.chronics_handler.sample_next_chronics([0.25, 0.75])
                obs = env.reset()
                print(obs.month)
                # it prints "2" with probability 0.75 and "8" with probability 0.25

        """
        self._prev_cache_id = -1
        if probabilities is None:
            probabilities = np.ones(self._order.shape[0])
        try:
            probabilities = np.array(probabilities, dtype=dt_float)
        except Exception as exc_:
            raise ChronicsError("Impossible to convert the probablities given to an array of float") from exc_
        
        sum_prob = probabilities.sum()
        if abs(sum_prob) <= 1e-5:
            raise ChronicsError("Impossible to use the given probabilities argument, it sums to 0. (or close enough to it)")
            
        # make sure it sums to 1
        probabilities /= sum_prob
        # take one at "random" among these
        selected = self.space_prng.choice(self._order, p=probabilities)
        id_sel = np.where(self._order == selected)[0]
        self._prev_cache_id = selected - 1
        return id_sel

    def reset(self):
        """
        Rebuilt the :attr:`Multifolder._order`. This should be called after a call to :func:`Multifolder.set_filter`
        is performed.

        .. warning:: This "reset" is different from the `env.reset`. It should be only called after the function to set
            the filtering function has been called.

            This "reset" only reset which chronics are used for the environment.

        Returns
        -------
        new_order: ``numpy.ndarray``, dtype: str
            The selected chronics paths after a call to this method.

        Notes
        -----
        Except explicitly mentioned, for example by :func:`Multifolder.set_filter` you should not use this
        function. This will erased every selection of chronics, every shuffle etc.

        """
        self._order = []
        self._prev_cache_id = 0
        for i, path in enumerate(self.subpaths):
            if not self._filter(path):
                continue
            self._order.append(i)

        if len(self._order) == 0:
            raise ChronicsError(
                'Impossible to initialize the Multifolder. Your "filter_fun" filters out all the '
                "possible scenarios."
            )
        self._order = np.array(self._order)
        return self.subpaths[self._order]

    def initialize(
        self,
        order_backend_loads,
        order_backend_prods,
        order_backend_lines,
        order_backend_subs,
        names_chronics_to_backend=None,
    ):

        self._order_backend_loads = order_backend_loads
        self._order_backend_prods = order_backend_prods
        self._order_backend_lines = order_backend_lines
        self._order_backend_subs = order_backend_subs
        self._names_chronics_to_backend = names_chronics_to_backend

        self.n_gen = len(order_backend_prods)
        self.n_load = len(order_backend_loads)
        self.n_line = len(order_backend_lines)

        if self._order is None:
            # initialize the cache
            self.reset()

        id_scenario = self._order[self._prev_cache_id]
        this_path = self.subpaths[id_scenario]
        self.data = self.gridvalueClass(
            time_interval=self.time_interval,
            sep=self.sep,
            path=this_path,
            max_iter=self.max_iter,
            chunk_size=self.chunk_size,
            **self._kwargs
        )
        if self.seed is not None:
            max_int = np.iinfo(dt_int).max
            seed_chronics = self.space_prng.randint(max_int)
            self.data.seed(seed_chronics)

        self.data.initialize(
            order_backend_loads,
            order_backend_prods,
            order_backend_lines,
            order_backend_subs,
            names_chronics_to_backend=names_chronics_to_backend,
        )

    def done(self):
        """
        Tells the :class:`grid2op.Environment` if the episode is over.

        Returns
        -------
        res: ``bool``
            Whether or not the episode, represented by :attr:`MultiFolder.data` is over.

        """
        return self.data.done()

    def load_next(self):
        """
        Load the next data from the current episode. It loads the next time step for the current episode.

        Returns
        -------
        See the return type of  :class:`GridStateFromFile.load_next` (or of :attr:`MultiFolder.gridvalueClass` if it
        has been changed) for more information.

        """
        return self.data.load_next()

    def check_validity(self, backend):
        """
        This method check that the data loaded can be properly read and understood by the :class:`grid2op.Backend`.

        Parameters
        ----------
        backend: :class:`grid2op.Backend`
            The backend used for the experiment.

        Returns
        -------
        See the return type of  :class:`GridStateFromFile.check_validity` (or of :attr:`MultiFolder.gridvalueClass` if it
        has been changed) for more information.
        """
        return self.data.check_validity(backend)

    def forecasts(self):
        """
        The representation of the forecasted grid state(s), if any.

        Returns
        -------
        See the return type of  :class:`GridStateFromFile.forecasts` (or of :attr:`MultiFolder.gridvalueClass` if it
        has been changed) for more information.
        """
        return self.data.forecasts()

    def tell_id(self, id_num, previous=False):
        """
        This tells this chronics to load for the next episode.
        By default, if id_num is greater than the number of episode, it is equivalent at restarting from the first
        one: episode are played indefinitely in the same order.

        Parameters
        ----------
        id_num: ``int`` | ``str``
            Id of the chronics to load.

        previous:
            Do you want to set to the previous value of this one or not (note that in general you want to set to
            the previous value, as calling this function as an impact only after `env.reset()` is called)
        """
        import pdb

        if isinstance(id_num, str):
            # new accepted behaviour starting 1.6.4
            # new in version 1.6.5: you only need to specify the chronics folder id and not the full path
            found = False
            for internal_id_, number in enumerate(self._order):
                if (
                    self.subpaths[number] == id_num
                    or os.path.join(self.path, id_num) == self.subpaths[number]
                ):
                    self._prev_cache_id = internal_id_
                    found = True

            if not found:
                raise ChronicsError(
                    f'Impossible to find the chronics with id "{id_num}". The call to '
                    f"`env.chronics_handler.tell_id(...)` cannot be performed."
                )
        else:
            # default behaviour prior to 1.6.4
            self._prev_cache_id = id_num
            self._prev_cache_id %= len(self._order)

        if previous:
            self._prev_cache_id -= 1
            self._prev_cache_id %= len(self._order)

    def get_id(self) -> str:
        """
        Full absolute path of the current folder used for the current episode.

        Returns
        -------
        res: ``str``
            Path from which the data are generated for the current episode.

        """
        return self.subpaths[self._order[self._prev_cache_id]]

    def max_timestep(self):
        return self.data.max_timestep()

    def shuffle(self, shuffler=None):
        """
        This method is used to have a better control on the order in which the subfolder containing the episode are
        processed.

        It can focus the evaluation on one specific folder, shuffle the folders, use only a subset of them etc. See the
        examples for more information.

        Parameters
        ----------
        shuffler: ``object``
            Shuffler should be a function that is called on :attr:`MultiFolder.subpaths` that will shuffle them.
            It can also be used to remove some path if needed (see example).

        Returns
        --------
        new_order: ``numpy.ndarray``, dtype: str
            The order in which the chronics will be looped through

        Examples
        ---------
        If you want to simply shuffle the data you can do:

        .. code-block:: python

            # create an environment
            import numpy as np
            import grid2op
            env_name = "l2rpn_case14_sandbox"
            env = grid2op.make(env_name)

            # shuffle the chronics (uniformly at random, without duplication)
            env.chronics_handler.shuffle()
            # use the environment as you want, here do 10 episode with the selected data
            for i in range(10):
                obs = env.reset()
                print(f"Path of the chronics used: {env.chronics_handler.data.path}")
                done = False
                while not done:
                    act = ...
                    obs, reward, done, info = env.step(act)

            # re shuffle them (still uniformly at random, without duplication)
            env.chronics_handler.shuffle()

            # use the environment as you want, here do 10 episode with the selected data
            for i in range(10):
                obs = env.reset()
                print(f"Path of the chronics used: {env.chronics_handler.data.path}")
                done = False
                while not done:
                    act = ...
                    obs, reward, done, info = env.step(act)


        If you want to use only a subset of the path, say for example the path with index 1, 5, and 6

        .. code-block:: python

            # create an environment
            import numpy as np
            import grid2op
            env_name = "l2rpn_case14_sandbox"
            env = grid2op.make(env_name)

            # select the chronics (here 5 at random amongst the 10 "last" chronics of the environment)
            nb_chron = len(env.chronics_handler.chronics_used)
            chron_id_to_keep = np.random.choice(np.arange(nb_chron - 10, nb_chron), size=5, replace=True)
            env.chronics_handler.shuffle(lambda x: chron_id_to_keep)

            # use the environment as you want, here do 10 episode with the selected data
            for i in range(10):
                obs = env.reset()
                print(f"Path of the chronics used: {env.chronics_handler.data.path}")
                done = False
                while not done:
                    act = ...
                    obs, reward, done, info = env.step(act)

            # re shuffle them (uniformly at random, without duplication, among the chronics "selected" above.)
            env.chronics_handler.shuffle()

            # use the environment as you want, here do 10 episode with the selected data
            for i in range(10):
                obs = env.reset()
                print(f"Path of the chronics used: {env.chronics_handler.data.path}")
                done = False
                while not done:
                    act = ...
                    obs, reward, done, info = env.step(act)

        .. warning:: Though it is possible to use this "shuffle" function to only use some chronics, we highly
            recommend you to have a look at the sections :ref:`environment-module-chronics-info` or
            :ref:`environment-module-train-val-test`. It is likely that you will find better way to do
            what you want to do there. Use this last example with care then.

        .. warning:: As stated on the :func:`MultiFolder.reset`, any call to `env.chronics_handler.reset`
            will remove anything related to shuffling, including the selection of chronics !

        """
        if shuffler is None:

            def shuffler(x):
                return x[self.space_prng.choice(len(x), size=len(x), replace=False)]

        self._order = 1 * shuffler(self._order)
        return self.subpaths[self._order]

    @property
    def chronics_used(self):
        """return the full path of the chronics currently in use."""
        return self.subpaths[self._order]

    def set_chunk_size(self, new_chunk_size):
        self.chunk_size = new_chunk_size

    def split_and_save(self, datetime_beg, datetime_end, path_out):
        """
        This function allows you to split the data (keeping only the data between datetime_beg and datetime_end) and to
        save it on your local machine. This is espacially handy if you want to extract only a piece of the dataset we
        provide for example.

        Parameters
        ----------
        datetime_beg: ``dict``
            Keys are the name id of the scenarios you want to save. Values
            are the corresponding starting date and time (in "%Y-%m-ùd %H:%M"
            format). See example for more information.
        datetime_end: ``dict``
            keys must be the same as in the "datetime_beg" argument.

            See example for more information

        path_out: ``str``
            The path were the data will be stored.

        Examples
        ---------

        Here is a short example on how to use it

        .. code-block:: python

            import grid2op
            import os
            env = grid2op.make()

            env.chronics_handler.real_data.split_and_save({"004": "2019-01-08 02:00",
                                                 "005": "2019-01-30 08:00",
                                                 "006": "2019-01-17 00:00",
                                                 "007": "2019-01-17 01:00",
                                                 "008": "2019-01-21 09:00",
                                                 "009": "2019-01-22 12:00",
                                                 "010": "2019-01-27 19:00",
                                                 "011": "2019-01-15 12:00",
                                                 "012": "2019-01-08 13:00",
                                                 "013": "2019-01-22 00:00"},
                                                {"004": "2019-01-11 02:00",
                                                 "005": "2019-02-01 08:00",
                                                 "006": "2019-01-18 00:00",
                                                 "007": "2019-01-18 01:00",
                                                 "008": "2019-01-22 09:00",
                                                 "009": "2019-01-24 12:00",
                                                 "010": "2019-01-29 19:00",
                                                 "011": "2019-01-17 12:00",
                                                 "012": "2019-01-10 13:00",
                                                 "013": "2019-01-24 00:00"},
                                      path_out=os.path.join("/tmp"))

        """
        if not isinstance(datetime_beg, dict):
            datetime_beg_orig = datetime_beg
            datetime_beg = {}
            for subpath in self.subpaths:
                id_this_chron = os.path.split(subpath)[-1]
                datetime_beg[id_this_chron] = datetime_beg_orig
        if not isinstance(datetime_end, dict):
            datetime_end_orig = datetime_end
            datetime_end = {}
            for subpath in self.subpaths:
                id_this_chron = os.path.split(subpath)[-1]
                datetime_end[id_this_chron] = datetime_end_orig
        seed_chronics_all = {}
        for subpath in self.subpaths:
            id_this_chron = os.path.split(subpath)[-1]
            if not id_this_chron in datetime_beg:
                continue
            tmp = self.gridvalueClass(
                time_interval=self.time_interval,
                sep=self.sep,
                path=subpath,
                max_iter=self._max_iter,
                chunk_size=self.chunk_size,
            )
            seed_chronics = None
            if self.seed is not None:
                max_int = np.iinfo(dt_int).max
                seed_chronics = self.space_prng.randint(max_int)
                tmp.seed(seed_chronics)
            seed_chronics_all[subpath] = seed_chronics
            tmp.initialize(
                self._order_backend_loads,
                self._order_backend_prods,
                self._order_backend_lines,
                self._order_backend_subs,
                self._names_chronics_to_backend,
            )
            path_out_chron = os.path.join(path_out, id_this_chron)
            tmp.split_and_save(
                datetime_beg[id_this_chron], datetime_end[id_this_chron], path_out_chron
            )

            meta_params = {}
            meta_params["datetime_beg"] = datetime_beg
            meta_params["datetime_end"] = datetime_end
            meta_params["path_out"] = path_out
            meta_params["all_seeds"] = seed_chronics_all
            try:
                with open(
                    os.path.join(path_out, "split_and_save_meta_params.json"),
                    "w",
                    encoding="utf-8",
                ) as f:
                    json.dump(obj=meta_params, fp=f, sort_keys=True, indent=4)
            except Exception as exc_:
                warnings.warn(
                    'Impossible to save the "metadata" for the chronics with error:\n"{}"'
                    "".format(exc_)
                )
                
    def fast_forward(self, nb_timestep):
        self.data.fast_forward(nb_timestep)



# Copyright (c) 2019-2021, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import warnings
import copy
import numpy as np

from grid2op.dtypes import dt_int
from grid2op.Opponent import BaseOpponent
from grid2op.Exceptions import OpponentError


class GeometricOpponent(BaseOpponent):
    """
    This opponent will disconnect lines randomly among the attackable lines `lines_attacked`.
    The sampling is done according to the lines load factor (ratio <current going through the line> to <thermal limit
    of the line>)
    (see init for more details).

    The time of the attack is sampled according to a geometric distribution
    """

    def __init__(self, action_space):
        BaseOpponent.__init__(self, action_space)
        self._do_nothing = None
        self._attacks = None
        self._lines_ids = None
        self._next_attack_time = None
        self._attack_hazard_rate = None
        self._recovery_minimum_duration = None
        self._recovery_rate = None
        self._pmax_pmin_ratio = None
        self._attack_times = None
        self._attack_waiting_times = None
        self._attack_durations = None
        self._attack_counter = None
        self._number_of_attacks = None
        self._episode_max_time = None
        self._env = None  # I need to keep a pointer to the environment for computing the maximum length of the episode
        # this is the constructor:
        # it should have the exact same signature as here

    def init(
        self,
        partial_env,
        lines_attacked=(),
        attack_every_xxx_hour=24,
        average_attack_duration_hour=4,
        minimum_attack_duration_hour=2,
        pmax_pmin_ratio=4,
        **kwargs,
    ):
        """
        Generic function used to initialize the derived classes. For example, if an opponent reads from a file, the
        path where is the file is located should be pass with this method.

        Parameters
        ----------
        partial_env: grid2op Environment
            A pointer to the environment that initializes the opponent

        lines_attacked: ``list``
            The list of lines that the XPOpponent should be able to disconnect

        attack_every_xxx_hour: ``float``
            Provide the average duration between two attacks. Note that this should be greater
            than `average_attack_duration_hour` as, for now, an agent can only do one consecutive attack.
            You should provide it in "number of hours" and not in "number of steps"

            It is used to compute the `attack_hazard_rate`.
            Attacks time are sampled with a duration distribution. For this opponent, we use the simplest of these
            distributions : The geometric disribution
            https://en.wikipedia.org/wiki/Geometric_distribution (the discrete time counterpart of the exponential
            distribution).
            The attack_hazard_rate is the main parameter of this distribution. It can be seen as the (constant)
            probability of having an attack
            in the next step. It is also the inverse of the expectation of the time to an attack.

        average_attack_duration_hour: ``float``
            Give, in number of hours, the average attack duration. This should be greater than
            `recovery_minimum_duration_hour`

            Used to compute the `recovery_rate`:
            Recovery times are random or at least should have a random part.
            In our case, we will say that the recovery time is equal to a fixed time (safety procedure time) plus a
            random time (investigations
            and repair operations) sampled according to a geometric distribution

        minimum_attack_duration_hour: ``int``
            Minimum duration of an attack (give it in hour)

        pmax_pmin_ratio: ``float``
            Ratio between the probability of the most likely line to be disconnected and the least likely one.
        """
        self._env = partial_env

        if len(lines_attacked) == 0:
            warnings.warn(
                f"The opponent is deactivated as there is no information as to which line to attack. "
                f'You can set the argument "kwargs_opponent" to the list of the line names you want '
                f' the opponent to attack in the "make" function.'
            )

        # Store attackable lines IDs
        self._lines_ids = []
        for l_name in lines_attacked:
            l_id = np.where(self.action_space.name_line == l_name)
            if len(l_id) and len(l_id[0]):
                self._lines_ids.append(l_id[0][0])
            else:
                raise OpponentError(
                    'Unable to find the powerline named "{}" on the grid. For '
                    "information, powerlines on the grid are : {}"
                    "".format(l_name, sorted(self.action_space.name_line))
                )

        # Pre-build attacks actions
        self._do_nothing = self.action_space({})
        self._attacks = []
        for l_id in self._lines_ids:
            a = self.action_space({"set_line_status": [(l_id, -1)]})
            self._attacks.append(a)
        self._attacks = np.array(self._attacks)

        # Opponent's attack and recovery rates and minimum duration
        # number of steps per hour
        ts_per_hour = 3600.0 / partial_env.delta_time_seconds
        self._recovery_minimum_duration = int(
            minimum_attack_duration_hour * ts_per_hour
        )
        if average_attack_duration_hour < minimum_attack_duration_hour:
            raise OpponentError(
                "The average duration of an attack cannot be lower than the minimum time of an attack"
            )
        elif average_attack_duration_hour == minimum_attack_duration_hour:
            raise OpponentError(
                "Case average_attack_duration_hour == minimum_attack_duration_hour is not supported "
                "at the moment"
            )
        self._recovery_rate = 1.0 / (
            ts_per_hour * (average_attack_duration_hour - minimum_attack_duration_hour)
        )
        if attack_every_xxx_hour <= average_attack_duration_hour:
            raise OpponentError(
                "attack_every_xxx_hour <= average_attack_duration_hour is not supported at the moment."
            )
        self._attack_hazard_rate = 1.0 / (
            ts_per_hour * (attack_every_xxx_hour - average_attack_duration_hour)
        )

        # Opponent's pmax pmin ratio
        self._pmax_pmin_ratio = pmax_pmin_ratio

        # Episode max time
        self._episode_max_time = self._get_episode_duration()

        # Sample attack times and durations for the whole episode
        self.sample_attack_times_and_durations()

        # Set the attack counter to 0
        self._attack_counter = 0

    def _get_episode_duration(self):
        tmp = self._env.max_episode_duration()
        if (not np.isfinite(tmp)) or (tmp == np.iinfo(tmp).max):
            raise OpponentError(
                "Geometric opponent only works (for now) with a known finite episode duration."
            )
        return tmp

    def reset(self, initial_budget):
        # Sample attack times and durations for the whole episode
        self.sample_attack_times_and_durations()

        # Reset the attack counter to 0
        self._attack_counter = 0

        self._next_attack_time = None

        # Episode max time
        self._episode_max_time = self._get_episode_duration()

    def sample_attack_times_and_durations(self):
        self._attack_times = []
        self._attack_waiting_times = []
        self._attack_durations = []
        self._number_of_attacks = 0

        t = 0  # t=0 at the beginning of the episode

        while t < self._episode_max_time:
            # Sampling the next time to attack
            t_to_attack = self.space_prng.geometric(p=self._attack_hazard_rate)
            t_of_attack = t + t_to_attack
            t = t_of_attack
            if t < self._episode_max_time:
                self._attack_waiting_times.append(t_to_attack)
                self._attack_times.append(t_of_attack)
                self._number_of_attacks += 1
                # Sampling the attack duration
                attack_duration = (
                    self._recovery_minimum_duration
                    + self.space_prng.geometric(p=self._recovery_rate)
                )
                self._attack_durations.append(attack_duration)
                t = t + attack_duration
            # TODO : Log these times and durations in a log.file

        self._attack_times = np.array(self._attack_times).astype(dt_int)
        self._attack_waiting_times = np.array(self._attack_waiting_times).astype(dt_int)
        self._attack_durations = np.array(self._attack_durations).astype(dt_int)

    def tell_attack_continues(self, observation, agent_action, env_action, budget):
        self._next_attack_time = None

    def attack(self, observation, agent_action, env_action, budget, previous_fails):
        """
        This method is the equivalent of "attack" for a regular agent.
        Opponent, in this framework can have more information than a regular agent (in particular it can
        view time step t+1), it has access to its current budget etc.
        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The last observation (at time t)
        opp_reward: ``float``
            THe opponent "reward" (equivalent to the agent reward, but for the opponent) TODO do i add it back ???
        done: ``bool``
            Whether the game ended or not TODO do i add it back ???
        agent_action: :class:`grid2op.Action.Action`
            The action that the agent took
        env_action: :class:`grid2op.Action.Action`
            The modification that the environment will take.
        budget: ``float``
            The current remaining budget (if an action is above this budget, it will be replaced by a do nothing.
        previous_fails: ``bool``
            Wheter the previous attack failed (due to budget or ambiguous action)
        Returns
        -------
        attack: :class:`grid2op.Action.Action`
            The attack performed by the opponent. In this case, a do nothing, all the time.
        duration: ``int``
            The duration of the attack (if ``None`` then the attack will be made for the longest allowed time)
        """
        # During creation of the environment, do not attack
        if observation is None:
            return None, None

        # If there are no more attacks to come, do not attack
        if self._attack_counter >= self._number_of_attacks:
            return None, None

        if previous_fails:
            # i cannot do the attack, it failed (so self._attack_counter >= 1)
            self._next_attack_time = (
                self._attack_waiting_times[self._attack_counter]
                + self._attack_durations[self._attack_counter - 1]
            )

        # Set the time of the next attack
        if self._next_attack_time is None:
            self._next_attack_time = (
                1 + self._attack_waiting_times[self._attack_counter]
            )

        attack_duration = self._attack_durations[self._attack_counter]
        self._next_attack_time -= 1

        # If the attack time has not come yet, do not attack
        if self._next_attack_time > 0:
            return None, None
        else:
            # Attack is launched
            self._attack_counter += 1

        # If all attackable lines are disconnected, abort attack
        status = observation.line_status[self._lines_ids]
        if np.all(status == False):
            return None, None

        available_attacks = self._attacks[status]

        # If we have a unique attackable line we just attack it
        if len(available_attacks) == 1:
            return available_attacks[0], attack_duration

        # We have several lines, so we need to choose one
        # This will be according to their load factor (rho)
        rho = observation.rho[self._lines_ids][status]

        # The rho_rank vector is the ranking of the lines according
        # to their rho (load factor)
        # 0 : for the line with the lowest load factor
        # (n_attackable_lines - 1) : for the line with the highest load factor
        temp = rho.argsort()
        rho_ranks = np.empty_like(temp)
        rho_ranks[temp] = np.arange(len(rho))

        # We choose the attacked line using a Boltzmann distribution
        # on rho ranks, with a beta parameter (temperature) set to ensure
        # that the probability ratio between the most and the least prefered
        # lines is equal to the pmax_pmin_ratio parameter
        n_attackable_line = len(available_attacks)
        b_beta = np.log(self._pmax_pmin_ratio) / (n_attackable_line - 1)
        raw_probabilities = np.exp(b_beta * rho_ranks)
        b_probabilities = raw_probabilities / raw_probabilities.sum()
        attack = self.space_prng.choice(available_attacks, p=b_probabilities)
        return attack, attack_duration

    def get_state(self):
        return (
            self._attack_times,
            self._attack_waiting_times,
            self._attack_durations,
            self._number_of_attacks,
        )

    def set_state(self, my_state):
        (
            _attack_times,
            _attack_waiting_times,
            _attack_durations,
            _number_of_attacks,
        ) = my_state
        self._attack_times = 1 * _attack_times
        self._attack_waiting_times = 1 * _attack_waiting_times
        self._attack_durations = 1 * _attack_durations
        self._number_of_attacks = 1 * _number_of_attacks

    def _custom_deepcopy_for_copy(self, new_obj, dict_=None):
        super()._custom_deepcopy_for_copy(new_obj, dict_)
        if dict_ is None:
            raise OpponentError("Impossible to deep copy an Opponent without a pointer "
                                "to the original env, named `partial_env`.")

        new_obj._attacks = copy.deepcopy(self._attacks)
        new_obj._lines_ids = copy.deepcopy(self._lines_ids)
        new_obj._next_attack_time = copy.deepcopy(self._next_attack_time)
        new_obj._attack_hazard_rate = copy.deepcopy(self._attack_hazard_rate)
        new_obj._recovery_minimum_duration = copy.deepcopy(
            self._recovery_minimum_duration
        )
        new_obj._recovery_rate = copy.deepcopy(self._recovery_rate)
        new_obj._pmax_pmin_ratio = copy.deepcopy(self._pmax_pmin_ratio)
        new_obj._attack_counter = copy.deepcopy(self._attack_counter)
        new_obj._episode_max_time = copy.deepcopy(self._episode_max_time)
        new_obj._env = dict_[
            "partial_env"
        ]  # I need to keep a pointer to the environment for computing the maximum length of the episode



from grid2op.Action import TopologyAndDispatchAction
from grid2op.Reward import RedispReward
from grid2op.Rules import DefaultRules
from grid2op.Chronics import Multifolder
from grid2op.Chronics import GridStateFromFileWithForecasts
from grid2op.Backend import PandaPowerBackend

config = {
    "backend": PandaPowerBackend,
    "action_class": TopologyAndDispatchAction,
    "observation_class": None,
    "reward_class": RedispReward,
    "gamerules_class": DefaultRules,
    "chronics_class": Multifolder,
    "grid_value_class": GridStateFromFileWithForecasts,
    "volagecontroler_class": None,
    "thermal_limits": [
        3.84900179e02,
        3.84900179e02,
        2.28997102e05,
        2.28997102e05,
        2.28997102e05,
        1.52664735e04,
        2.28997102e05,
        3.84900179e02,
        3.84900179e02,
        1.83285800e02,
        3.84900179e02,
        3.84900179e02,
        2.28997102e05,
        2.28997102e05,
        6.93930612e04,
        3.84900179e02,
        3.84900179e02,
        2.40562612e02,
        3.40224266e03,
        3.40224266e03,
    ],
    "names_chronics_to_grid": None,
}



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

__all__ = [
    "BaseGymAttrConverter",
    "GymEnv",
    "GymObservationSpace",
    "GymActionSpace",
    "ScalerAttrConverter",
    "MultiToTupleConverter",
    "ContinuousToDiscreteConverter",
    "BoxGymObsSpace",
    "BoxGymActSpace",
    "MultiDiscreteActSpace",
    "DiscreteActSpace",
]

from grid2op.gym_compat.utils import _MAX_GYM_VERSION_RANDINT, GYM_VERSION, GYMNASIUM_AVAILABLE, GYM_AVAILABLE

if GYM_AVAILABLE is False and GYMNASIUM_AVAILABLE is False:
    raise ImportError("Neither gymnasium nor gym are installed. The `grid2op.gym_compat` module cannot be used.")

# base for all gym converter
from grid2op.gym_compat.base_gym_attr_converter import BaseGymAttrConverter
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.base_gym_attr_converter import BaseGymnasiumAttrConverter
    __all__.append("BaseGymnasiumAttrConverter")
if GYM_AVAILABLE:
    from grid2op.gym_compat.base_gym_attr_converter import BaseLegacyGymAttrConverter
    __all__.append("BaseLegacyGymAttrConverter")

# the environment (by default with dict encoding)
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.gymenv import GymnasiumEnv
    __all__.append("GymnasiumEnv")
if GYM_AVAILABLE:
    from grid2op.gym_compat.gymenv import GymEnv_Legacy, GymEnv_Modern
    __all__.append("GymEnv_Legacy")
    __all__.append("GymEnv_Modern")

# define the default env to use
if GYMNASIUM_AVAILABLE:
    GymEnv = GymnasiumEnv
else:
    if GYM_VERSION <= _MAX_GYM_VERSION_RANDINT:
        GymEnv = GymEnv_Legacy
    else:
        GymEnv = GymEnv_Modern
    
# action space (as Dict)
from grid2op.gym_compat.gym_act_space import GymActionSpace
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.gym_act_space import GymnasiumActionSpace
    __all__.append("GymnasiumActionSpace")
if GYM_AVAILABLE:
    from grid2op.gym_compat.gym_act_space import LegacyGymActionSpace
    __all__.append("LegacyGymActionSpace")
    
# observation space (as Dict)
from grid2op.gym_compat.gym_obs_space import GymObservationSpace
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.gym_obs_space import GymnasiumObservationSpace
    __all__.append("GymnasiumObservationSpace")
if GYM_AVAILABLE:
    from grid2op.gym_compat.gym_obs_space import LegacyGymObservationSpace
    __all__.append("LegacyGymObservationSpace")
    
from grid2op.gym_compat.scaler_attr_converter import ScalerAttrConverter
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.scaler_attr_converter import ScalerAttrConverterGymnasium
    __all__.append("ScalerAttrConverterGymnasium")
if GYM_AVAILABLE:
    from grid2op.gym_compat.scaler_attr_converter import ScalerAttrConverterLegacyGym
    __all__.append("ScalerAttrConverterLegacyGym")


from grid2op.gym_compat.multi_to_tuple_converter import MultiToTupleConverter
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.multi_to_tuple_converter import MultiToTupleConverterGymnasium
    __all__.append("MultiToTupleConverterGymnasium")
if GYM_AVAILABLE:
    from grid2op.gym_compat.multi_to_tuple_converter import MultiToTupleConverterLegacyGym
    __all__.append("MultiToTupleConverterLegacyGym")
    

from grid2op.gym_compat.continuous_to_discrete import ContinuousToDiscreteConverter
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.continuous_to_discrete import ContinuousToDiscreteConverterGymnasium
    __all__.append("ContinuousToDiscreteConverterGymnasium")
if GYM_AVAILABLE:
    from grid2op.gym_compat.continuous_to_discrete import ContinuousToDiscreteConverterLegacyGym
    __all__.append("ContinuousToDiscreteConverterLegacyGym")
   
# observation space (as Box)
from grid2op.gym_compat.box_gym_obsspace import BoxGymObsSpace
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.box_gym_obsspace import BoxGymnasiumObsSpace
    __all__.append("BoxGymnasiumObsSpace")
if GYM_AVAILABLE:
    from grid2op.gym_compat.box_gym_obsspace import BoxLegacyGymObsSpace
    __all__.append("BoxLegacyGymObsSpace")


from grid2op.gym_compat.box_gym_actspace import BoxGymActSpace
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.box_gym_actspace import BoxGymnasiumActSpace
    __all__.append("BoxGymnasiumActSpace")
if GYM_AVAILABLE:
    from grid2op.gym_compat.box_gym_actspace import BoxLegacyGymActSpace
    __all__.append("BoxLegacyGymActSpace")


from grid2op.gym_compat.multidiscrete_gym_actspace import MultiDiscreteActSpace
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.multidiscrete_gym_actspace import MultiDiscreteActSpaceGymnasium
    __all__.append("MultiDiscreteActSpaceGymnasium")
if GYM_AVAILABLE:
    from grid2op.gym_compat.multidiscrete_gym_actspace import MultiDiscreteActSpaceLegacyGym
    __all__.append("MultiDiscreteActSpaceLegacyGym")
    

from grid2op.gym_compat.discrete_gym_actspace import DiscreteActSpace
if GYMNASIUM_AVAILABLE:
    from grid2op.gym_compat.discrete_gym_actspace import DiscreteActSpaceGymnasium
    __all__.append("DiscreteActSpaceGymnasium")
if GYM_AVAILABLE:
    from grid2op.gym_compat.discrete_gym_actspace import DiscreteActSpaceLegacyGym
    __all__.append("DiscreteActSpaceLegacyGym")


# TODO doc and test




# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import os
from abc import ABC, abstractmethod
from grid2op.Space import RandomObject
from grid2op.Observation import BaseObservation
from grid2op.Action import BaseAction, ActionSpace


class BaseAgent(RandomObject, ABC):
    """
    This class represents the base class of an BaseAgent. All bot / controller / agent used in the Grid2Op simulator
    should derived from this class.

    To work properly, it is advise to create BaseAgent after the :class:`grid2op.Environment` has been created and reuse
    the :attr:`grid2op.Environment.Environment.action_space` to build the BaseAgent.

    Attributes
    -----------
    action_space: :class:`grid2op.Action.ActionSpace`
        It represent the action space ie a tool that can serve to create valid action. Note that a valid action can
        be illegal or ambiguous, and so lead to a "game over" or to a error. But at least it will have a proper size.

    """

    def __init__(self, action_space: ActionSpace):
        RandomObject.__init__(self)
        self.action_space = copy.deepcopy(action_space)

    def reset(self, obs: BaseObservation):
        """
        This method is called at the beginning of a new episode.
        It is implemented by agents to reset their internal state if needed.

        Attributes
        -----------
        obs: :class:`grid2op.Observation.BaseObservation`
            The first observation corresponding to the initial state of the environment.
        """
        pass

    def seed(self, seed: int) -> None:
        """
        This function is used to guarantee that the "pseudo random numbers" generated and used by the agent instance
        will be deterministic.

        This guarantee, if the recommendation in :func:`BaseAgent.act` are followed that the agent will produce the same
        set of actions if it faces the same observations in the same order. This is particularly important for
        random agent.

        You can override this function with the method of your choosing, but if you do so, don't forget to call
        `super().seed(seed)`.

        Parameters
        ----------
        seed: ``int``
            The seed used

        Returns
        -------
        seed: ``tuple``
            a tuple of seed used
        """

        return super().seed(seed), self.action_space.seed(seed)

    @abstractmethod
    def act(self, observation: BaseObservation, reward: float, done : bool=False) -> BaseAction:
        """
        This is the main method of an BaseAgent. Given the current observation and the current reward (ie the reward
        that the environment send to the agent after the previous action has been implemented).

        Notes
        -----
        In order to be reproducible, and to make proper use of the
        :func:`BaseAgent.seed` capabilities, you must absolutely NOT use the `random` python module (which will not
        be seeded) nor the `np.random` module and avoid any other "sources" of pseudo random numbers.

        You can adapt your code the following way. Instead of using `np.random` use `self.space_prng`.

        For example, if you wanted to write
        `np.random.randint(1,5)` replace it by `self.space_prng.randint(1,5)`. It is the same for `np.random.normal()`
        that is
        replaced by `self.space_prng.normal()`.

        You have an example of such usage in :func:`RandomAgent.my_act`.

        If you really need other sources of randomness (for example if you use tensorflow or torch) we strongly
        recommend you to overload the :func:`BaseAgent.seed` accordingly. In that

        Parameters
        ----------
        observation: :class:`grid2op.Observation.BaseObservation`
            The current observation of the :class:`grid2op.Environment.Environment`

        reward: ``float``
            The current reward. This is the reward obtained by the previous action

        done: ``bool``
            Whether the episode has ended or not. Used to maintain gym compatibility

        Returns
        -------
        res: :class:`grid2op.Action.PlaybleAction`
            The action chosen by the bot / controler / agent.

        """
        return self.action_space()

    def save_state(self, savestate_path :os.PathLike):  
        """
        An optional method to save the internal state of your agent.
        The saved state can later be re-loaded with `self.load_state`, e.g. to repeat 
        a Grid2Op time step with exactly the same internal parameterization. This
        can be useful to repeat Grid2Op experiments and analyze why your agent performed 
        certain actions in past time steps. Concept developed by Fraunhofer IEE KES.

        Notes
        -----
        First, the internal state your agent consists of attributes that are contained in 
        the :class:`grid2op.Agent.BaseAgent` and :class:`grid2op.Agent.BaseAgent.action_space`.
        Examples are the parameterization and seeds of the random number generator that your
        agent uses. Such attributes can easily be obtained with the :func:`getattr` and stored
        in a common file format, such as `.npy`.
        
        Second, your agent may contain custom attributes, such as e.g. a vector of line indices 
        from a Grid2Op observation. You could obtain and save them in the same way as explained 
        before.

        Third, your agent may contain very specific modules such as `Tensorflow` that
        do not support the simple :func:`getattr`. However, these modules normally have
        their own methods to save an internal state. Examples of such methods are
        :func:`save_weights` that you can integrate in your implementation of `self.save_state`.
        
        Parameters
        ----------
        savestate_path: ``string``
            The path to which your agent state variables should be saved
        """
        pass
    
    def load_state(self, loadstate_path :os.PathLike):  
        """
        An optional method to re-load the internal agent state that was saved with `self.save_state`. 
        This can be useful to re-set your agent to an earlier simulation time step and reproduce 
        past experiments with Grid2Op. Concept developed by Fraunhofer IEE KES.

        Notes
        -----
        First, the internal state your agent consists of attributes that are contained in 
        the :class:`grid2op.Agent.BaseAgent` and :class:`grid2op.Agent.BaseAgent.action_space`.
        Such attributes can easily be re-set with :func:`setattr`.
        
        Second, your agent may contain custom attributes, such as e.g. a vector of line indices 
        from a Grid2Op observation. You can re-set them with :func:`setattr` as well.

        Third, your agent may contain very specific modules such as `Tensorflow` that
        do not support the simple :func:`setattr`. However, these modules normally have
        their own methods to re-load an internal state. Examples of such methods are
        :func:`load_weights` that you can integrate in your implementation of `self.load_state`.
        
        Parameters
        ----------
        savestate_path: ``string``
            The path from which your agent state variables should be loaded
        """
        pass



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
from itertools import chain
import warnings
import copy

from grid2op.Rules.BaseRules import BaseRules
from grid2op.Rules.LookParam import LookParam
from grid2op.Rules.PreventReconnection import PreventReconnection
from grid2op.Rules.PreventDiscoStorageModif import PreventDiscoStorageModif
from grid2op.Exceptions import (
    IllegalAction, Grid2OpException
)


class RulesByArea(BaseRules):
    """
    This subclass combine :class:`PreventReconnection`, :class: `PreventDiscoStorageModif` to be applied on the whole grid at once,
    while a specifique method look for the legality of simultaneous actions taken on defined areas of a grid.
    An action is declared legal if and only if:

      - It doesn't reconnect more power lines than what is stated in the actual game _parameters
        :class:`grid2op.Parameters`
      - It doesn't attempt to act on more substations and lines within each area that what is stated in the actual game _parameters
        :class:`grid2op.Parameters`
      - It doesn't attempt to modify the power produce by a turned off storage unit
      
    Example
    ---------
    If you want the environment to take into account the rules by area, you can achieve it with:

    .. code-block:
        import grid2op
        from grid2op.Rules.rulesByArea import RulesByArea
        
        # First you set up the areas within the RulesByArea class
        my_gamerules_byarea = RulesByArea([[0,1,2,3,4,5,6,7],[8,9,10,11,12,13,14]])
        # Then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "l2rpn_case14_sandbox"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,gamerules_class=my_gamerules_byarea)

    """

    def __init__(self, areas_list):
        """
        The initialization of the rule with a list of list of ids of substations composing the aimed areas.
        Parameters
        ----------
        areas_list : list of areas, each placeholder containing the ids of substations of each defined area
        """
        if isinstance(areas_list, list):
            self.substations_id_by_area = {i : sorted(k) for i, k in enumerate(areas_list)}
        elif isinstance(areas_list, dict):
            self.substations_id_by_area = {i : copy.deepcopy(k) for i, k in areas_list.items()}
        else:
            raise Grid2OpException("Impossible to create a rules when area_list is neither a list nor a dict")
        needs_cleaning = False
        for area_nm, area_subs in self.substations_id_by_area.items():
            if not np.array_equal(np.unique(area_subs), area_subs):
                warnings.warn(f"There are duplicate substation for area {area_nm}")
                needs_cleaning = True
        
        if needs_cleaning:
            self.substations_id_by_area = {i : np.unique(k) for i, k in self.substations_id_by_area.items()}

    def initialize(self, env):
        """
        This function is used to inform the class instance about the environment 
        specification and check no substation of the grid are left ouside an area. 
        Parameters
        ----------
        env: :class:`grid2op.Environment.Environment`
            An environment instance properly initialized.

        """
        n_sub = env.n_sub
        n_sub_rule = np.sum([len(set(list_ids)) for list_ids in self.substations_id_by_area.values()])
        if n_sub_rule != n_sub: 
            raise Grid2OpException("The number of listed ids of substations in rule initialization does not match the number of "
                                   "substations of the chosen environement. Look for missing ids or doublon")
        else:
            self.lines_id_by_area = {key : sorted(list(chain(*[[item for item in np.where(env.line_or_to_subid == subid)[0]
                                    ] for subid in subid_list]))) for key,subid_list in self.substations_id_by_area.items()}


    def __call__(self, action, env):
        """
        See :func:`BaseRules.__call__` for a definition of the _parameters of this function.
        """
        is_legal, reason = PreventDiscoStorageModif.__call__(self, action, env)
        if not is_legal:
            return False, reason
            
        is_legal, reason = self._lookparam_byarea(action, env)
        if not is_legal:
            return False, reason
            
        return PreventReconnection.__call__(self, action, env)
        

    def can_use_simulate(self, nb_simulate_call_step, nb_simulate_call_episode, param):
        return LookParam.can_use_simulate(
            self, nb_simulate_call_step, nb_simulate_call_episode, param
        )
    
    def _lookparam_byarea(self, action, env):
        """
        See :func:`BaseRules.__call__` for a definition of the parameters of this function.
        """
        # at first iteration, env.current_obs is None...
        powerline_status = env.get_current_line_status()

        aff_lines, aff_subs = action.get_topological_impact(powerline_status)
        if any([(aff_lines[line_ids]).sum() > env._parameters.MAX_LINE_STATUS_CHANGED for line_ids in self.lines_id_by_area.values()]):
            ids = [[k for k in np.where(aff_lines)[0] if k in line_ids] for line_ids in self.lines_id_by_area.values()]
            return False, IllegalAction(
                "More than {} line status affected by the action in one area: {}"
                "".format(env.parameters.MAX_LINE_STATUS_CHANGED, ids)
            )
        if any([(aff_subs[sub_ids]).sum() > env._parameters.MAX_SUB_CHANGED for sub_ids in self.substations_id_by_area.values()]):
            ids = [[k for k in np.where(aff_subs)[0] if k in sub_ids] for sub_ids in self.substations_id_by_area.values()]
            return False, IllegalAction(
                "More than {} substation affected by the action in one area: {}"
                "".format(env.parameters.MAX_SUB_CHANGED, ids)
            )
        return True, None
 


# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import os
import sys
import numpy as np
from datetime import datetime
import warnings

from grid2op.Episode import EpisodeData
from grid2op.Exceptions import Grid2OpException
from grid2op.Plot.PlotPyGame import PlotPyGame
from grid2op.Exceptions.PlotExceptions import PyGameQuit

try:
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
    import pygame

    can_plot = True
except Exception as e:
    can_plot = False
    pass

try:
    # from array2gif import write_gif
    import imageio
    import imageio_ffmpeg

    can_save_gif = True
except:
    can_save_gif = False


class EpisodeReplay(object):
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

        Prefer using the class `grid2op.Episode.EpisodeReplay`

    This class allows to see visually what an agent has done during an episode. It uses for now the "PlotPygame" as the
    method to plot the different states of the system. It reads directly data from the runner.

    It can be used the following manner.

    .. code-block:: python

        import grid2op
        agent_class = grid2op.Agent.DoNothingAgent  # change that for studying other agent
        env = grid2op.make()  # make the default environment
        runner = grid2op.Runner.Runner(**env.get_params_for_runner(), agentClass=agent_class)
        path_log = "agent_log"  # where the runner will output the standardized data when running the agent.
        res = runner.run(nb_episode=1, path_save=path_log)

        # and when it's done, you can visualize it this way:
        episode_replay = EpisodeReplay(agent_path=path_log)
        episode_id = res[0][1]
        episode_replay.plot_episode(episode_id, max_fps=10)

        # you can pause by clicking the "space" key
        # At any time, you can quit by pressing the "esc" key or the "exit" button of the window.

    Attributes
    ----------
    agent_path: ``str``
        The path were the log of the agent are stored. It is recommended to use a :class:`grid2op.Runner.Runner`
        to save tha log of the agent.

    episode_data: :class:`grid2op.EpisodeData.EpisodeData`, optional
        The last data of the episode inspected.
    """

    def __init__(self, agent_path):
        warnings.warn(
            "This whole class has been deprecated. Use `grid2op.PlotGrid module instead`",
            category=DeprecationWarning,
        )

        if not os.path.exists(agent_path):
            raise Grid2OpException(
                'Nothing is found at "{}" where an agent path should have been.'.format(
                    agent_path
                )
            )
        self.agent_path = agent_path
        self.episode_data = None

        if not can_save_gif:
            warnings.warn(
                'The final video will not be saved as "imageio" and "imageio_ffmpeg" packages cannot be '
                'imported. Please try "{} -m pip install imageio imageio-ffmpeg"'.format(
                    sys.executable
                )
            )

    def replay_episode(self, episode_id, max_fps=10, video_name=None, display=True):
        """
        .. warning:: /!\\\\ This class is deprecated /!\\\\

            Prefer using the class `grid2op.Episode.EpisodeReplay`

        When called, this function will start the display of the episode in a "mini movie" format.

        Parameters
        ----------
        episode_id: ``str``
            ID of the episode to replay

        max_fps: ``int``
            Maximum "frame per second". When it's low, you will have more time to look at each frame, but the episode
            will last longer. When it's high, episode will be faster, but frames will stay less time on the screen.

        video_name: ``str``
            In beta mode for now. This allows to save the "video" of the episode in a gif or a mp4 for example.

        Returns
        -------

        """
        path_ep = os.path.join(self.agent_path, episode_id)
        if not os.path.exists(path_ep):
            raise Grid2OpException(
                'No episode is found at "{}" where the episode should have been.'.format(
                    path_ep
                )
            )
        if video_name is None:
            if not can_save_gif:
                raise Grid2OpException(
                    'The final video cannot be saved as "imageio" and "imageio_ffmpeg" '
                    "packages cannot be imported. Please try "
                    '"{} -m pip install imageio imageio-ffmpeg"'.format(sys.executable)
                )

        self.episode_data = EpisodeData.from_disk(
            agent_path=self.agent_path, name=episode_id
        )
        plot_runner = PlotPyGame(
            self.episode_data.observation_space, timestep_duration_seconds=1.0 / max_fps
        )
        nb_timestep_played = int(self.episode_data.meta["nb_timestep_played"])
        all_obs = [el for el in self.episode_data.observations]
        all_reward = [el for el in self.episode_data.rewards]
        if video_name is not None:
            total_array = np.zeros(
                (
                    nb_timestep_played + 1,
                    plot_runner.video_width,
                    plot_runner.video_height,
                    3,
                ),
                dtype=np.uint8,
            )

        if display is False:
            plot_runner.deactivate_display()

        for i, (obs, reward) in enumerate(zip(all_obs, all_reward)):
            timestamp = datetime(
                year=obs.year,
                month=obs.month,
                day=obs.day,
                hour=obs.hour_of_day,
                minute=obs.minute_of_hour,
            )
            try:
                plot_runner.plot_obs(
                    observation=obs,
                    reward=reward,
                    timestamp=timestamp,
                    done=i == nb_timestep_played - 1,
                )
                array_ = pygame.surfarray.array3d(plot_runner.screen)
                if video_name is not None:
                    total_array[i, :, :, :] = array_.astype(np.uint8)
            except PyGameQuit:
                break

        if video_name is not None:
            imageio.mimwrite(video_name, np.swapaxes(total_array, 1, 2), fps=max_fps)
        plot_runner.close()



# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import os
import numpy as np
from typing import Optional, Tuple
from grid2op.Space import RandomObject
from datetime import timedelta, datetime


# TODO logger !
class BaseHandler(RandomObject):
    """This is the base class that represents a time series "handler".
    
    .. versionadded:: 1.9.0
    
    Each "handler" will be reponsible to produce the data for "one single type of elements" 
    of the grid. For example you will have 1 handler for "load_p", one for "load_q", another
    one for "load_p_forecasted" etc. This allows some great flexibility to the way you want 
    to retrieve data, but can be quite verbose as every type of data needs to be "handled".
    
    A "handler" will, for a certain type of data (*eg* load_p or maintenance etc.)
    handle the way this data type is generated.
    
    To be a valid "handler" an class must first inherit from :class:`BaseHandler` and 
    implements (for all types of handlers):
    
    - :func:`BaseHandler.initialize` : to initialize the handler from the environment 
      data (number of loads, lines etc.)
    - :func:`BaseHandler.done` : whether or not this "handler" is over or not.
    - :func:`BaseHandler.check_validity` : check if the input data are valid with the backend, 
      for example if you read from a csv
      number of columns should match number of element
    - :func:`BaseHandler.next_chronics` : called just before the start of a scenario.
    
    If the data represents "real time" data (*ie* the data seen by the agent in real 
    time in the observation) then it needs also to implement:
    
    - :func:`BaseHandler.load_next` : to "generate" data for the next steps
    
    If the data represents "forecast data" (*ie* the data accessed by the agent when it uses
    :func:`grid2op.Observation.BaseObservation.simulate` or :class:`grid2op.simulator.Simulator`
    or :func:`grid2op.Observation.BaseObservation.get_forecast_env`) then it needs to implement:
    
    - :func:`BaseHandler.forecast` : to retrieve the forecast at a given horizon
    
    And if the "handler" represents maintenance data, then it needs to implement:
    
    - :func:`BaseHandler.load_next` : that returns a boolean vector for whether or not 
      each powerline is in maintenance
    - :func:`BaseHandler.load_next_maintenance` : to "generate" data for the next steps
    
    .. seealso:: 
        
        :class:`grid2op.Chronics.FromHandlers` which is the things that "consumes" the handlers to output 
        the data read by the :class:`grid2op.Environment.Environment`
    
    """
    def __init__(self, array_name, max_iter=-1, h_forecast=(5, )):
        super().__init__()
        self.max_iter : int = max_iter
        self.init_datetime : Optional[datetime] = None
        self.time_interval : Optional[timedelta] = None
        self.array_name : str = array_name
        self._h_forecast : tuple = copy.deepcopy(h_forecast)
        self.path : Optional[os.PathLike] = None
        self.max_episode_duration : Optional[int] = None
    
    def set_max_iter(self, max_iter: Optional[int]) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        The the maximum number of iteration this handler is able to produce.
        
        `-1` means "forever". This is set by the :class:`grid2op.Chronics.FromHandlers`

        Parameters
        ----------
        max_iter : Optional[int]
            Maximum number of iterations
            
        """
        if max_iter is not None:
            self.max_iter = int(max_iter)
        else:
            self.max_iter = -1
            
    def set_max_episode_duration(self, max_episode_duration: Optional[int]) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        The the maximum number of iteration the environment will use. It is a way
        to synchronize all the handlers for the same environment

        Parameters
        ----------
        max_episode_duration : Optional[int]
            Maximum number of iterations for the current grid2op environment
            
        """
        if max_episode_duration is not None:
            self.max_episode_duration = int(max_episode_duration)
        else:
            self.max_episode_duration = None
        
    def get_max_iter(self) -> int:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        The maximum number of iterations this particular handler can generate.
        
        `-1` means "forever" otherwise it should be a > 0 integers.

        Returns
        -------
        int
            The maximum number of iterations this particular handler can generate.
        """
        return self.max_iter
    
    def set_path(self, path: os.PathLike) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This method is used by the :class:`grid2op.Chronics.FromHandlers` to inform this handler about the location
        where the required data for this handler could be located.

        Parameters
        ----------
        path : os.PathLike
            The path to look for the data
        """
        self.path = path
    
    def set_chunk_size(self, chunk_size: Optional[int]) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        optional: when data are read from the hard drive (*eg* :class:`grid2op.Chronics.handlers.CSVHandler`)
        this can inform the handler about the number of data to proceed at each 'step'.
        
        .. note::
           Do not use this function directly, it should be used only from the environment.
           
        .. seealso::
            This can be set by a call to `env.chronics_handler.set_chunk_size(chunk_size)`

        Parameters
        ----------
        chunk_size : Optional[int]
            The desired chunk size
        """
        # Chunk size is part of public API but has no sense for 
        # data not read from a disk
        pass
        
    def set_times(self,
                  init_datetime : datetime,
                  time_interval : timedelta) -> None:

        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This method is used by the :class:`grid2op.Chronics.FromHandlers` to inform this handler 
        about the intial datetime of the episode and the duration between two steps.
        
        .. note::
           Do not use this function directly, it should be used only from the environment.
           
        Parameters
        ----------
        init_datetime : datetime
            The initial datetime.
            
        time_interval : timedelta
            The time between two steps of the environment.
            
        """
        self.init_datetime = init_datetime
        self.time_interval = time_interval
    
    def _clear(self):
        self.init_datetime = None
        self.time_interval = None

    def get_kwargs(self, dict_ : dict) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        If some extra parameters are needed to build your handler "from scratch" you should copy them here. This is used
        when creating a runner for example.

        Parameters
        ----------
        dict_ : dict
            The dictionnary to update with the parameters.
        """
        # no need to remember special kwargs for the base class
        pass
    
    def set_h_forecast(self, h_forecast : Tuple[int]) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This method is used by the :class:`grid2op.Chronics.FromHandlers` to inform this handler 
        about the different forecast horizons available.

        .. seealso::
            :func:`BaseHandler.get_available_horizons`
            
        Parameters
        ----------
        h_forecast : Tuple[int]
            A tuple containing the different forecast horizons available. The horizons should be given in minutes,
            for example `handler.set_h_forecast((5, 10))` tells this handler that forecasts are available for 5 and 
            10 minutes ahead.
            
        """
        self._h_forecast = copy.deepcopy(h_forecast)
    
    def get_available_horizons(self) -> Tuple:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This methods returns the available forecast horizons (in minutes) known by this handler.
        
        .. seealso::
            :func:`BaseHandler.set_h_forecast`
            
        Returns
        -------
        Tuple
            A tuple containing the different forecast horizons available. The horizons should be given in minutes,
            for example `handler.set_h_forecast((5, 10))` tells this handler that forecasts are available for 5 and 
            10 minutes ahead.
            
        """
        return copy.deepcopy(self._h_forecast)
        
    def initialize(self,
                   order_backend_arrays,
                   names_chronics_to_backend) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This function is called by the :class:`grid2op.Chronics.FromHandlers` after the current handler has been created.
        Its goal is to initialize it with the relevant data from the environment.
        
        For example, if this handler represents "load_p" then `order_backend_arrays` will be the name of 
        each load in the environment and `names_chronics_to_backend` is a dictionnary mapping the name in the
        data to the names as read by the grid simulator / the backend.

        Parameters
        ----------
        order_backend_arrays : np.ndarray
            numpy array representing the name of the element in the grid
            
        names_chronics_to_backend : dict
            mapping between the names in `order_backend_arrays` and the names found in the data.
            
        """
        raise NotImplementedError()
    
    def done(self) -> bool:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        Whether or not this handler has done generating the data. It can be "done" in the case it 
        reads data from a csv and you are at the bottom line of the csv for example.

        Returns
        -------
        bool
            Whether it is "done" or not.
        """
        raise NotImplementedError()
    
    def load_next(self, dict_: dict) -> Optional[np.ndarray]:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
        
        This function is called by the :class:`grid2op.Chronics.FromHandlers`
        
        Used by the environment handlers ("load_p", "load_q", "prod_p" and "prod_v" only). When this function
        is called, it should return the next state of the type of data it is responsible for. If the previous
        state should not be modified, then this function can returns "None".
        
        This is called exactly once per step.

        Parameters
        ----------
        dict_ : dict
            A dictionnary representing the other "previous" data type. This function is always called in the same order:
            
            1) on "load_p"
            2) on "load_q"
            3) on "gen_p"
            4) on "gen_v"
            
            So if your handler is reponsible for "gen_p" then this dictionnary might contain 2 items:
            
            - key: "load_p", value: all the active loads for the environment at the same step (if the values are modified
              by the relevant handlers)
            - key: "load_q", value: all the reactive laods for the environment at the same step (if the values are modified
              by the relevant handlers)

        Returns
        -------
        Optional[np.ndarray]
            The new values (if any)
        """
        raise NotImplementedError()
    
    def check_validity(self, backend):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This function is called by the :class:`grid2op.Chronics.FromHandlers` after all the handlers have
        been initialized.
        
        Its role is to make sure that every handlers can "handle" the data of the environment smoothly. 
        
        It is called after each "env.reset()" call.

        Parameters
        ----------
        backend : :class:`grid2op.Backend.Backend`
            The backend used in the environment.
        """
        raise NotImplementedError()    
    
    def load_next_maintenance(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This function is used only if the handler is reponsible for "maintenance". It is called
        exactly once per step.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            maintenance time: np.ndarray
                Time for next maintenance for each powerline
            
            maintenance duration: np.ndarray
                Duration of the next maintenance, for each powerline
                
        """
        raise NotImplementedError()
    
    def load_next_hazard(self) -> np.ndarray:
        # TODO
        raise NotImplementedError()
        
    def forecast(self,
                 forecast_horizon_id : int,
                 inj_dict_env : dict,
                 inj_dict_previous_forecast : dict,
                 # eg gen_p_handler if this is set to gen_p_for_handler:
                 env_handler : "BaseHandler",  
                 # list of the 4 env handlers: (load_p_handler, load_q_handler, gen_p_handler, gen_v_handler)
                 env_handlers : Tuple["BaseHandler", "BaseHandler", "BaseHandler", "BaseHandler"]
                 ) -> Optional[np.ndarray] :
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This function is called by the :class:`grid2op.Chronics.FromHandlers` only for the handlers responsible 
        for some "forecasts", which are "load_p_forecasted", "load_q_forecasted", "prod_p_forecasted", 
        "prod_v_forecasted".
        
        It is called exactly once per step and per horizon.

        It's similar to :func:`BaseHandler.load_next` with different inputs (because "forecast" are more complicated that 
        just real time data)
        
        .. seealso:: :func:`BaseHandler.load_next`
        
        Parameters
        ----------
        forecast_horizon_id : int
            The `id` of the horizon concerns. The horizon id is the index of the current horizon in the list :attr:`BaseHandler._h_forecast`
            
        inj_dict_env : dict
            The dictionnary containing the data of the environment (not the forecast) if data have been modified by the relevant handlers.
            
        inj_dict_previous_forecast : dict
            Similar to the `dict_` parameters of :func:`BaseHandler.load_next`
            
        env_handler : BaseHandler
            The handler of the same type as this one, but for the environment.
            
            For example, if this handler deals with "`load_q_forecasted`" then 
            `env_handler` will be the handler of `load_q`.
            
        env_handlers : Tuple[:class:`BaseHandler`, :class:`BaseHandler`, :class:`BaseHandler`, :class:`BaseHandler`]
            In these you have all the environment handlers in a tuple.
            
            The order is: "load_p", "load_q", "prod_p", "prod_v".

        Returns
        -------
        Optional[np.ndarray]
            The forecast (in the shape of numpy array) or None if nothing should be returned.
        """
        raise NotImplementedError()
    
    def get_future_data(self, horizon: int, quiet_warnings : bool=False) -> Optional[np.ndarray]:
        """        
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
        
        This function is for example used in the 
        :class:`grid2op.Chronics.handlers.PerfectForecastHandler`: to generate a 
        "perfect forecast" this class will use this function to "have a look"
        into the future through this function.
        
        This function is for example implemented in 
        :class:`grid2op.Chronics.handlers.CSVHandler`

        Parameters
        ----------
        horizon : int
            The horizon (in minutes) to which we want the data.

        quiet_warnings: bool
            Whether to issue a warning (default, if quiet_warnings is False) or not
            
        Returns
        -------
        Optional[np.ndarray]
            The data that will be generated  in `horizon` minutes.
        """
        return None
    
    def next_chronics(self) -> None:
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
            
        This function is called by the :class:`grid2op.Chronics.FromHandlers` at the
        end of each episode when the next episode is loaded.
        """
        return None



# Copyright (c) 2019-2021, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.


class DoNothingLog:
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    A class to emulate the behaviour of a logger, but that does absolutely nothing.
    """

    INFO = 2
    WARNING = 1
    ERROR = 0

    def __init__(self, max_level=2):
        self.max_level = max_level

    def warn(self, *args, **kwargs):
        pass

    def info(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass

    def warning(self, *args, **kwargs):
        pass


class ConsoleLog(DoNothingLog):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    A class to emulate the behaviour of a logger, but that prints on the console
    """

    def __init__(self, max_level=2):
        DoNothingLog.__init__(self, max_level)

    def warn(self, *args, **kwargs):
        if self.max_level >= self.WARNING:
            if args:
                print('WARNING: "{}"'.format(", ".join(args)))
            if kwargs:
                print("WARNING: {}".format(kwargs))

    def info(self, *args, **kwargs):
        if self.max_level >= self.INFO:
            if args:
                print('INFO: "{}"'.format(", ".join(args)))
            if kwargs:
                print("INFO: {}".format(kwargs))

    def error(self, *args, **kwargs):
        if self.max_level >= self.ERROR:
            if args:
                print('ERROR: "{}"'.format(", ".join(args)))
            if kwargs:
                print("ERROR: {}".format(kwargs))

    def warning(self, *args, **kwargs):
        if self.max_level >= self.WARNING:
            if args:
                print('WARNING: "{}"'.format(", ".join(args)))
            if kwargs:
                print("WARNING: {}".format(kwargs))



__all__ = [
    "RulesChecker",
    "DefaultRules",
    "AlwaysLegal",
    "BaseRules",
    "LookParam",
    "PreventReconnection",
    "PreventDiscoStorageModif",
    "RulesByArea",
]

from grid2op.Rules.RulesChecker import RulesChecker
from grid2op.Rules.DefaultRules import DefaultRules
from grid2op.Rules.AlwaysLegal import AlwaysLegal
from grid2op.Rules.BaseRules import BaseRules
from grid2op.Rules.LookParam import LookParam
from grid2op.Rules.PreventReconnection import PreventReconnection
from grid2op.Rules.PreventDiscoStorageModif import PreventDiscoStorageModif
from grid2op.Rules.rulesByArea import RulesByArea
import warnings


class LegalAction(BaseRules):
    def __init__(self, *args, **kwargs):
        BaseRules.__init__(self, *args, **kwargs)
        warnings.warn(
            'LegalAction class has been renamed "BaseRules". '
            "This class LegalAction will be removed in future versions.",
            category=PendingDeprecationWarning,
        )


class GameRules(RulesChecker):
    def __init__(self, *args, **kwargs):
        RulesChecker.__init__(self, *args, **kwargs)
        warnings.warn(
            'GameRules class has been renamed "RulesChecker". '
            "This class GameRules will be removed in future versions.",
            category=PendingDeprecationWarning,
        )


class PreventReconection(PreventReconnection):
    def __init__(self, *args, **kwargs):
        PreventReconection.__init__(self, *args, **kwargs)
        warnings.warn(
            'PreventReconection class has been renamed "PreventReconnection". '
            "This class Action will be removed in future versions.",
            category=PendingDeprecationWarning,
        )



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.gym_compat.utils import GYM_AVAILABLE, GYMNASIUM_AVAILABLE
from grid2op.gym_compat.utils import check_gym_version


class __AuxBaseGymAttrConverter(object):
    """
    TODO work in progress !

    Need help if you can :-)
    
    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`BaseGymAttrConverter` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`BaseGymnasiumAttrConverter`), otherwise it will
          inherit from gym (and will be exactly :class:`BaseLegacyGymAttrConverter`)
        - :class:`BaseGymnasiumAttrConverter` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`BaseLegacyGymAttrConverter` will inherit from gym if it's available and never from
          from gymnasium
        
        See :ref:`gymnasium_gym` for more information
        
    """

    def __init__(self, space=None, gym_to_g2op=None, g2op_to_gym=None):
        check_gym_version(type(self)._gymnasium)
        self.__is_init_super = (
            False  # is the "super" class initialized, do not modify in child class
        )

        self._is_init_space = False  # is the instance initialized
        self._my_gym_to_g2op = None
        self._my_g2op_to_gym = None
        self.my_space = None

        if space is not None:
            self.base_initialize(space, gym_to_g2op, g2op_to_gym)

    def base_initialize(self, space, gym_to_g2op, g2op_to_gym):
        if self.__is_init_super:
            return
        self._my_gym_to_g2op = gym_to_g2op
        self._my_g2op_to_gym = g2op_to_gym
        self.my_space = space
        self._is_init_space = True
        self.__is_init_super = True

    def is_init_space(self):
        return self._is_init_space

    def initialize_space(self, space):
        if self._is_init_space:
            return
        if not isinstance(space, type(self)._SpaceType):
            raise RuntimeError(
                "Impossible to scale a converter if this one is not from type space.Space"
            )
        self.my_space = space
        self._is_init_space = True

    def gym_to_g2op(self, gym_object):
        """
        Convert a gym object to a grid2op object

        Parameters
        ----------
        gym_object:
            An object (action or observation) represented as a gym "ordered dictionary"

        Returns
        -------
        The same object, represented as a grid2op.Action.BaseAction or grid2op.Observation.BaseObservation.

        """
        if self._my_gym_to_g2op is None:
            raise NotImplementedError(
                "Unable to convert gym object to grid2op object with this converter"
            )
        return self._my_gym_to_g2op(gym_object)

    def g2op_to_gym(self, g2op_object):
        """
        Convert a gym object to a grid2op object

        Parameters
        ----------
        g2op_object:
            An object (action or observation) represented as a grid2op.Action.BaseAction or
            grid2op.Observation.BaseObservation

        Returns
        -------
        The same object, represented as a gym "ordered dictionary"

        """
        if self._my_g2op_to_gym is None:
            raise NotImplementedError(
                "Unable to convert grid2op object to gym object with this converter"
            )
        return self._my_g2op_to_gym(g2op_object)


if GYM_AVAILABLE:
    from gym.spaces import Space as LegGymSpace
    BaseLegacyGymAttrConverter = type("BaseLegacyGymAttrConverter",
                                     (__AuxBaseGymAttrConverter, ),
                                     {"_SpaceType": LegGymSpace, 
                                      "_gymnasium": False,
                                      "__module__": __name__})
    BaseLegacyGymAttrConverter.__doc__ = __AuxBaseGymAttrConverter.__doc__
    BaseGymAttrConverter = BaseLegacyGymAttrConverter
        

if GYMNASIUM_AVAILABLE:
    from gymnasium.spaces import Space
    BaseGymnasiumAttrConverter = type("BaseGymnasiumAttrConverter",
                                     (__AuxBaseGymAttrConverter, ),
                                     {"_SpaceType": Space, 
                                      "_gymnasium": True,
                                      "__module__": __name__})
    BaseGymnasiumAttrConverter.__doc__ = __AuxBaseGymAttrConverter.__doc__
    BaseGymAttrConverter = BaseGymnasiumAttrConverter



__all__ = [
    "ConstantReward",
    "EconomicReward",
    "FlatReward",
    "IncreasingFlatReward",
    "L2RPNReward",
    "RedispReward",
    "BridgeReward",
    "CloseToOverflowReward",
    "DistanceReward",
    "GameplayReward",
    "LinesReconnectedReward",
    "LinesCapacityReward",
    "CombinedReward",
    "CombinedScaledReward",
    "RewardHelper",
    "BaseReward",
    "EpisodeDurationReward",
    "AlarmReward",
    "N1Reward",
    # TODO it would be better to have a specific package for this, but in the mean time i put it here
    "L2RPNSandBoxScore",
    "L2RPNWCCI2022ScoreFun",
    "AlertReward",
    "_AlarmScore",
    "_NewRenewableSourcesUsageScore",
    "_AlertCostScore",
    "_AlertTrustScore"
]

from grid2op.Reward.constantReward import ConstantReward
from grid2op.Reward.economicReward import EconomicReward
from grid2op.Reward.flatReward import FlatReward
from grid2op.Reward.increasingFlatReward import IncreasingFlatReward
from grid2op.Reward.l2RPNReward import L2RPNReward
from grid2op.Reward.redispReward import RedispReward
from grid2op.Reward.bridgeReward import BridgeReward
from grid2op.Reward.closeToOverflowReward import CloseToOverflowReward
from grid2op.Reward.distanceReward import DistanceReward
from grid2op.Reward.gameplayReward import GameplayReward
from grid2op.Reward.linesReconnectedReward import LinesReconnectedReward
from grid2op.Reward.linesCapacityReward import LinesCapacityReward
from grid2op.Reward.combinedReward import CombinedReward
from grid2op.Reward.combinedScaledReward import CombinedScaledReward
from grid2op.Reward.rewardHelper import RewardHelper
from grid2op.Reward.baseReward import BaseReward
from grid2op.Reward.l2RPNSandBoxScore import L2RPNSandBoxScore
from grid2op.Reward.episodeDurationReward import EpisodeDurationReward
from grid2op.Reward.alarmReward import AlarmReward
from grid2op.Reward._alarmScore import _AlarmScore
from grid2op.Reward.n1Reward import N1Reward
from grid2op.Reward.l2rpn_wcci2022_scorefun import L2RPNWCCI2022ScoreFun
from grid2op.Reward.alertReward import AlertReward
from grid2op.Reward._newRenewableSourcesUsageScore import _NewRenewableSourcesUsageScore
from grid2op.Reward._alertCostScore import _AlertCostScore
from grid2op.Reward._alertTrustScore import _AlertTrustScore


import warnings


class Reward(BaseReward):
    def __init__(self, *args, **kwargs):
        BaseReward.__init__(self, *args, **kwargs)
        warnings.warn(
            'Reward class has been renamed "BaseReward". '
            "This class Action will be removed in future versions.",
            category=PendingDeprecationWarning,
        )
